:- module(set_parameter,[vsc_set/1,
			 set/1,
			 set/2,
			 setting/2]).

:- meta_predicate vsc_set(:), set(:), set(:,+), setting(:,-).

:- dynamic user:sayu_min_recall_in_area/1.

user:sayu_min_recall_in_area(0.2).

:- multifile user:set_hook/1.
:- dynamic user:set_hook/1.

vsc_set(T) :-
	set(T).

set(TM) :-
	var(TM), !,
	throw(error(instantiation_error)).
set(TM) :-
	( user:set_hook(TM) -> true ; true),
	strip_module(TM,M,T),
	functor(T,F,A),
	functor(N,F,A),
	get_spec_module(M,F,M1),
	retractall(M1:N),
	assert(M1:T).

get_spec_module(_,F,M1) :-
	spec_module(M1, F), !.
get_spec_module(M,_,M).

spec_module(debug_write, verbosity).
spec_module(load_folds, sayu_cv_folds).
spec_module(cv_attributes, cvsayu_classifier_type).
spec_module(cv_attributes, cvsayu_class).
spec_module(cv_attributes, cvsayu_mode).
spec_module(cv_attributes, cvsayu_object).
spec_module(cv_attributes, cvsayu_scoring_method).
spec_module(cv_attributes, cvsayu_shuffle_training_set).
spec_module(sayu_attributes, sayu_classifier_type).
spec_module(sayu_attributes, sayu_classifier_options).
spec_module(sayu_attributes, sayu_n_of_evaluated_atts).
spec_module(sayu_attributes, sayu_scoring_method).
spec_module(sayu_attributes, sayu_use_aggs).
spec_module(score_attributes, scorer).
spec_module(score_attributes, min_bias).
spec_module(score_attributes, max_stdev).
spec_module(score_attributes, cmp_scores_howto).
spec_module(score_attributes, min_improvement_threshold).
spec_module(score_attributes, target).
spec_module(user, sayu_cur_file).
spec_module(user, sayu_dump_file).
spec_module(user, sayu_folds).
spec_module(user, sayu_min_cov).
spec_module(user, sayu_min_pos).
spec_module(user, sayu_min_recall_in_area).
spec_module(user, sayu_mode).
spec_module(user, sayu_prefix).
spec_module(user, sayu_prefix_clause).
spec_module(user, sayu_start_seeds).
spec_module(user, sayu_scoring_seeds).
spec_module(vista, agg_nodes).
spec_module(vista, invented_arity).
spec_module(vista, program).
spec_module(vista, search_assert).
spec_module(vista, split).
spec_module(vista, use_disjoin).
spec_module(vista, use_hierarchy).
spec_module(vista, max_step).
spec_module(vista_aggregation, aggregation_type).
spec_module(vista_beam, beam_size).
spec_module(vista_bestfirst, max_beam_size).
spec_module(vista_bottom_clause, i).
spec_module(vista_constants, constants_from_bottom_clauses).
spec_module(vista_constants, correlation_file).
spec_module(vista_constants, demand_constants).
spec_module(vista_constants, min_constants_from_bottom_clauses).
spec_module(vista_correlation, correlation).
spec_module(vista_correlation, min).
spec_module(vista_correlation, min_corr).
spec_module(vista_eval_node, minacc).
spec_module(vista_eval_node, minpos).
spec_module(vista_eval_node, nodes).
spec_module(vista_eval_node, search).
spec_module(vista_eval_node, search_mode).
spec_module(vista_eval_node, strict_clauses).
spec_module(vista_eval_node, dynamic_hierarchy).
spec_module(vista_foil, use_negation).
spec_module(vista_constants, min_constants_from_bottom_clauses).
spec_module(vista_hierarchy, hierarchy_stratified).
spec_module(weka_attributes, weka_scoring_method).
spec_module(weka_attributes, weka_classifier_type).
spec_module(weka_attributes, weka_classifier_options).
spec_module(weka_attributes, weka_n_of_evaluated_atts).

set(MName,Value) :-
	strip_module(MName,M,Name),
	atom(Name),
	T =.. [Name,Value],
	S =.. [Name,_],
	get_spec_module(M, Name, M1),
	retractall(M1:S),
	assert(M1:T).

setting(MName,Value) :-
	strip_module(MName,M,Name),
	atom(Name),
	T =.. [Name,Value],
	get_spec_module(M, Name, M1),
	call(M1:T).


