
:- module(vista_utils,[aleph_member/2,
		       average/2,
		       geometric_average/2,
		       average_and_stdev/3,
		       average2/3,
		       average_and_stdev2/4,
		       min/2,
		       min/3,
		       max/2,
		       max/3,
		       remove_dups/2,
		       clause_length/2,
		       clause_length/3,
		       run_goals/1,
		       list_to_conjunction/2,
		       conjunction_to_list/2,
		       randomize_list/2]).

:- use_module(random,[nextrandom/1]).

clause_length((_:-B),L) :- !,
	clause_length(B,0,L).
clause_length(_,0).

clause_length((A,B),I0,I) :- !,
	clause_length(A,I0,I1),
	clause_length(B,I1,I).
clause_length(_,I0,I) :-
	I is I0+1.

aleph_member(X,[X|_]).
aleph_member(X,[_|T]):-
	aleph_member(X,T).

randomize_list(Literals, RandomLiterals) :-
	gen_random(Literals,KeyLiterals),
	keysort(KeyLiterals,Sorted),
	extract_literals(Sorted, RandomLiterals).

gen_random([],[]).
gen_random([L|Literals],[R-L|KeyLiterals]) :-
	nextrandom(R),
	gen_random(Literals,KeyLiterals).

extract_literals([], []).
extract_literals([_-L|Sorted], [L|RandomLiterals]) :-
	extract_literals(Sorted, RandomLiterals).

from(I,I,_).
from(O,Min,Max) :- Min < Max,
	Min1 is Min+1,
	from(O,Min1,Max).

build_0list(0, []) :- !.
build_0list(N,[0|L]) :-
	N1 is N-1,
	build_0list(N1,L).

for(I,I,_).
for(I,Min,Max) :-
	Min < Max,
	Min1 is Min+1,
	for(I,Min1,Max).


%
% Removes duplicated elements in a list
%
remove_dups(Original, New) :-
	sort(Original, New).

average(L, Av) :-
	avg(L, 0, 0.0, Av).

avg([], N, Sum, Av) :-
	Av is Sum/N.
avg(A.Vs, N0, Sum0, Av) :-
	N1 is N0+1,
	Sum is Sum0+A,
	avg(Vs, N1, Sum, Av).

average_and_stdev(L, Av, StDev) :-
	avg(L, 0, 0.0, Av),
	stdev(L, Av, 0, 0.0, StDev).

average2(L1, L2, Av) :-
	avg2(L1, L2, 0, 0.0, Av).

average_and_stdev2(L1, L2, Av, StDev) :-
	avg2(L1, L2, 0, 0.0, Av),
	stdev2(L1, L2, Av, 0, 0.0, StDev).

avg2([], L, N, Sum, Av) :-
	avg(L, N, Sum, Av).
avg2(A.Vs, L, N0, Sum0, Av) :-
	N1 is N0+1,
	Sum is Sum0+A,
	avg2(Vs, L, N1, Sum, Av).

geometric_average(L, Avg) :-
	geo_avg(L, 0, 1.0, Avg).

geo_avg([], N, Prod, Av) :-
	Av is exp(Prod,1/N).
geo_avg(A.Vs, N0, Prod0, Av) :-
	N1 is N0+1,
	Prod is Prod0*A,
	avg(Vs, N1, Prod, Av).

stdev([], _, N, Sum, StDev) :-
	StDev is sqrt(Sum/(N-1)).
stdev(A.L, Av, N0, Dif0, StDev) :-
	N1 is N0+1,
	Dif1 is (A-Av)*(A-Av)+Dif0,
	stdev(L, Av, N1, Dif1, StDev).

stdev2([], L, Av, N, Sum, StDev) :-
	stdev(L, Av, N, Sum, StDev).
stdev2(A.L, L2, Av, N0, Dif0, StDev) :-
	N1 is N0+1,
	Dif1 is (A-Av)*(A-Av)+Dif0,
	stdev2(L, L2, Av, N1, Dif1, StDev).

min([Min0|L], Min) :-
	min(L, Min0, Min).

min([], Lik, Lik).
min([Lik|Liks], Lik0, LikF) :-
	( Lik  < Lik0 -> min(Liks, Lik, LikF) ; min(Liks, Lik0, LikF) ).

max([Max0|L], Max) :-
	max(L, Max0, Max).

max([], Lik, Lik).
max([Lik|Liks], Lik0, LikF) :-
	( Lik  > Lik0 -> max(Liks, Lik, LikF) ; max(Liks, Lik0, LikF) ).


list_to_conjunction([], true).
list_to_conjunction([H], H).
list_to_conjunction([H,H2|B], (H,NB)) :-
	list_to_conjunction([H2|B], NB).

conjunction_to_list(Conj, List) :-
	conjunction_to_list(Conj, List, []).

conjunction_to_list(true) --> [].
conjunction_to_list((A,B)) --> !,
	conjunction_to_list(A),
	conjunction_to_list(B).
conjunction_to_list(A) --> [A].

run_goals([]).
run_goals(G.Gs) :-
	call(user:G),
	run_goals(Gs).

