import sys, os

def readRuleFile(filename):
	f = open(filename, 'r')
	lines = f.readlines()
	f.close()
	ruleList = list()
	ruleBuilder = ''
	for line in lines:
		line = line.strip()
		if not line:
			continue
		ruleBuilder += ' ' + line
		# period is at the end of the rule
		if line.endswith('.'):
			# add it to our list, trim it since it will start with a space now
			ruleList.append(ruleBuilder.strip())
			# and prepare for the next one
			ruleBuilder = ''
	return ruleList

def readStratumSet(filename):
	f = open(filename, 'r')
	lines = f.readlines()
	f.close()
	retSet = set()
	for line in lines:
		line = line.strip()
		if not line:
			continue
		# get rid of the period at the end
		line = line.replace('.','')
		retSet.add(line)
	return retSet

def createAlephInput(rule, readAllPrefix):
	f = open(ALEPH_INPUT, 'w')
	f.write('read_all(\'' + readAllPrefix + '\').\n')
	f.write('rdhyp.\n')
	f.write(rule + '\n')
	f.write('tell(\'' + POS_COVERAGE + '\').\n')
	f.write('covers.\n')
	f.write('told.\n')
	f.write('tell(\'' + NEG_COVERAGE + '\').\n')
	f.write('coversn.\n')
	f.write('told.\n')
	f.write('halt.\n')
	f.close()

def readCoveredExamples(filename):
	f = open(filename, 'r')
	lines = f.readlines()
	f.close()
	retList = list()
	for line in lines:
		line = line.strip()
		if not line:
			continue
		# is it an example or some output garbage
		if line.endswith(').'):
			# append without the period for consistency
			retList.append(line.replace('.',''))
	return retList

def readCoverage(targetSet, otherSet):
	pos = readCoveredExamples(POS_COVERAGE)
	neg = readCoveredExamples(NEG_COVERAGE)
	# track the counts for each of the 4 groups
	pt = po = nt = no = 0
	# first do positives
	for ex in pos:
		if ex in targetSet:
			pt += 1
		elif ex in otherSet:
			po += 1
		else:
			print('Example ' + ex + ' not in target or other set')
	# then do negatives
	for ex in neg:
		if ex in targetSet:
			nt += 1
		elif ex in otherSet:
			no += 1
		else:
			print('Example ' + ex + ' not in target or other set')
	return (pt, po, nt, no)

def main(args):
	rules = readRuleFile(args[1])
	targetSet = readStratumSet(args[2])
	otherSet = readStratumSet(args[3])
	readAllPrefix = args[4]
	for rule in rules:
		createAlephInput(rule, readAllPrefix)
		# execute the Aleph script thing
		os.system('yap -l ~/Aleph/aleph.pl < ' + ALEPH_INPUT + ' &> /dev/null')
		# and count coverage
		print('Coverage for ' + rule)
		cov = readCoverage(targetSet, otherSet)
		print('pt=' + str(cov[0]) + ' po=' + str(cov[1]) + ' nt=' + str(cov[2]) + ' no=' + str(cov[3]))
		print('')
	# and clean up
	os.remove(ALEPH_INPUT)
	os.remove(POS_COVERAGE)
	os.remove(NEG_COVERAGE)

ALEPH_INPUT = 'aleph_input.tmp'
POS_COVERAGE = 'pos_cover.tmp'
NEG_COVERAGE = 'neg_cover.tmp'
main(sys.argv)