#!/bin/sh
#data folds will be in $FS and must be numbered from 1
#we assume two strata in differential sayu
FS0="$HOME/ILP/sayu/fs_strata0"
FS1="$HOME/ILP/sayu/fs_strata1"
# examples are in $FS/$EX$i.f and $FS/$EX$i.n
EX=OStest
#background file name is $BG.b
BG="train"
# number of nodes before sayu moves to a different seed
NODES=10000
# change to use your own YAP and Aleph
#1 start from empty network (initial sayu)
USE_AGGS=4
# naive bayes, tan, majority vote
#     public static final int NB = 0;
#     public static final int TAN = 1;
#     public static final int FULL_TAN = 3;
#     public static final int VOTE = 2;
CLASSIFIER_TYPE=1
# scoring method: accuracy, PR AUC, ROC AUC, conditional likelihood, etc.
#     public static final int ACC = 0;
#     public static final int PR_AREA = 1;
#     public static final int ROC_AREA = 2;
#     public static final int CLL = 3;
#     public static final int MSQE = 4;
#     public static final int X_ENT = 5;
#     public static final int LIFT_AREA = 6;
SCORING_METHOD=6
#
# default threshold for PR_AREA
#
MIN_THRESHOLD=0.2 #this isn't actually used for SAYL
SAYL_THRESHOLD=$1
MINPOS=$2
# YAP install in this system
YAP=/u/vitor/Yap/bins/java/yap
# SEARCH_MODE em, eager
SEARCH_MODE=$3
# directory where logs and theories will be stored
PREFIX=mammo_t"${SAYL_THRESHOLD/./}"_mp"$MINPOS"_"$SEARCH_MODE"
mkdir -p "$PREFIX"
for i in 1 2 3 4 5 6 7 8 9 10; do
  PROB_FILE="$PREFIX"/probs_"$i".yap
  IPREFIX="$PREFIX"_"$i"_
  IFILE="$FILE"fold"$i"  
  BK="$IPREFIX"_"$BG"
  CORRELATION_FILE=acorrs_"$i".yap
  ln -sf "$BG".b "$BK".b
  let "train1=($i+1) % 10 + 1"
  let "train2=($i+2) % 10 + 1"
  let "train3=($i+3) % 10 + 1"
  let "train4=($i+4) % 10 + 1"
  let "tune1=($i+5) % 10 + 1"
  let "tune2=($i+6) % 10 + 1"
  let "tune3=($i+7) % 10 + 1"
  let "tune4=($i+8) % 10 + 1"
  let "tune5=($i+9) % 10 + 1"
  let "test=($i) % 10 + 1"

  cat "$FS0"/"$EX""$train1".f \
      "$FS0"/"$EX""$train2".f \
      "$FS0"/"$EX""$train3".f \
      "$FS0"/"$EX""$train4".f > "$IPREFIX"train0.f
  cat "$FS0"/"$EX""$tune1".f \
      "$FS0"/"$EX""$tune2".f \
      "$FS0"/"$EX""$tune3".f \
      "$FS0"/"$EX""$tune4".f \
      "$FS0"/"$EX""$tune5".f > "$IPREFIX"tune0.f
  cat "$FS0"/"$EX""$test".f > "$IPREFIX"test0.f
  cat "$IPREFIX"train0.f \
      "$FS1"/"$EX""$train1".f \
      "$FS1"/"$EX""$train2".f \
      "$FS1"/"$EX""$train3".f \
      "$FS1"/"$EX""$train4".f > "$IPREFIX"train.f
  cat "$IPREFIX"tune0.f \
      "$FS1"/"$EX""$tune1".f \
      "$FS1"/"$EX""$tune2".f \
      "$FS1"/"$EX""$tune3".f \
      "$FS1"/"$EX""$tune4".f \
      "$FS1"/"$EX""$tune5".f > "$IPREFIX"tune.f
  cat "$IPREFIX"test0.f \
      "$FS1"/"$EX""$test".f > "$IPREFIX"test.f

  cat "$FS0"/"$EX""$train1".n \
      "$FS0"/"$EX""$train2".n \
      "$FS0"/"$EX""$train3".n \
      "$FS0"/"$EX""$train4".n > "$IPREFIX"train0.n
  cat "$FS0"/"$EX""$tune1".n \
      "$FS0"/"$EX""$tune2".n \
      "$FS0"/"$EX""$tune3".n \
      "$FS0"/"$EX""$tune4".n \
      "$FS0"/"$EX""$tune5".n > "$IPREFIX"tune0.n
  cat "$FS0"/"$EX""$test".n > "$IPREFIX"test0.n
  cat "$IPREFIX"train0.n \
      "$FS1"/"$EX""$train1".n \
      "$FS1"/"$EX""$train2".n \
      "$FS1"/"$EX""$train3".n \
      "$FS1"/"$EX""$train4".n > "$IPREFIX"train.n
  cat "$IPREFIX"tune0.n \
      "$FS1"/"$EX""$tune1".n \
      "$FS1"/"$EX""$tune2".n \
      "$FS1"/"$EX""$tune3".n \
      "$FS1"/"$EX""$tune4".n \
      "$FS1"/"$EX""$tune5".n > "$IPREFIX"tune.n
  cat "$IPREFIX"test0.n \
      "$FS1"/"$EX""$test".n > "$IPREFIX"test.n
  cat "$IPREFIX"train.f "$IPREFIX"tune.f > "$BK".f
  cat "$IPREFIX"train.n  "$IPREFIX"tune.n > "$BK".n
#cat <<EOF
nohup "$YAP" -l "$HOME"/ILP/sayu/aleph.yap  << EOF > "$PREFIX"/LOG.$i 2>&1 &
%yap_flag(gc_trace,verbose).
% saying which subset should be first thing,
read_all('$BK').
set(recordfile,'$PREFIX/$BG$i.trace').
vsc_set(sayu_dump_file('$PROB_FILE')).
vsc_set(sayu_min_recall_in_area($MIN_THRESHOLD)).
vsc_set(sayu_classifier_type($CLASSIFIER_TYPE)).
vsc_set(sayu_scoring_method($SCORING_METHOD)).
%vsc_set(sayu_start_seeds(200)).
vsc_set(sayu_prefix('$IPREFIX')).
set(nodes,$NODES).
set(verbosity,0).
set(minpos,$MINPOS).
vsc_set(sayu_mode($SEARCH_MODE)).
vsc_set(max_step(20)).
assert(user:sayl_threshold($SAYL_THRESHOLD)).
sayu_compute_strata0_sizes($IPREFIX).
Z is 3600, % 1 hour time limit
statistics(runtime,[T,_]), induce_sayu_timed(Z), statistics(runtime,[Tf,_]), D is Tf-T, format('ILP run took ~d msec.~n',[D]).


EOF
# rm -f "$BK".[bfn] "$IPREFIX"train.[fn] "$IPREFIX"tune.[fn] "$IPREFIX"test.[fn] "$IPREFIX"train0.[fn] "$IPREFIX"tune0.[fn] "$IPREFIX"test0.[fn]
done
