#!/bin/sh
# what fold and subfold is this setting for
FOLD=$1
SUB_FOLD=$2
#data folds will be in $FS and must be numbered from 1
#we assume two strata in differential sayu
FS0="$HOME/ILP/sayu/fs_strata0"
FS1="$HOME/ILP/sayu/fs_strata1"
# examples are in $FS/$EX$i.f and $FS/$EX$i.n
EX=OStest
#background file name is $BG.b
BG="train"
# number of nodes before sayu moves to a different seed
NODES=10000
# change to use your own YAP and Aleph
#1 start from empty network (initial sayu)
USE_AGGS=4
# naive bayes, tan, majority vote
#     public static final int NB = 0;
#     public static final int TAN = 1;
#     public static final int FULL_TAN = 3;
#     public static final int VOTE = 2;
CLASSIFIER_TYPE=1
# scoring method: accuracy, PR AUC, ROC AUC, conditional likelihood, etc.
#     public static final int ACC = 0;
#     public static final int PR_AREA = 1;
#     public static final int ROC_AREA = 2;
#     public static final int CLL = 3;
#     public static final int MSQE = 4;
#     public static final int X_ENT = 5;
#     public static final int LIFT_AREA = 6;
SCORING_METHOD=$6
SCORING_NAMES=('acc' 'pr' 'roc' 'cll' 'msqe' 'x_ent' 'lift')
SCORENAME=${SCORING_NAMES[$SCORING_METHOD]}
#
# default threshold for PR_AREA
#
MIN_THRESHOLD=0.2 #this isn't actually used for SAYL
SAYL_THRESHOLD=$3
# YAP install in this system
YAP=/u/vitor/Yap/bins/java/yap
# min pos cover
MINPOS=$4
# SEARCH_MODE em, eager
SEARCH_MODE=$5
# directory where logs and theories will be stored
PREFIX=mammo_icv_f"$FOLD"_sf"$SUB_FOLD"_t"${SAYL_THRESHOLD/./}"_mp"$MINPOS"_"$SCORENAME"_"$SEARCH_MODE"
mkdir -p "$PREFIX"
PROB_FILE="$PREFIX"/probs.yap
IPREFIX="$PREFIX"_
BK="$IPREFIX"_"$BG"
# this is how we would normally assign files to each fold
# we want to be consistent with the final runs
let "i=$FOLD"
ln -sf "$BG".b "$BK".b
let "train1=($i+1) % 10 + 1"
let "train2=($i+2) % 10 + 1"
let "train3=($i+3) % 10 + 1"
let "train4=($i+4) % 10 + 1"
let "tune1=($i+5) % 10 + 1"
let "tune2=($i+6) % 10 + 1"
let "tune3=($i+7) % 10 + 1"
let "tune4=($i+8) % 10 + 1"
let "tune5=($i+9) % 10 + 1"
let "test=($i) % 10 + 1"

# now we need to select the correct files for this subfold, avoding test
let "j=$SUB_FOLD"
ICV_ARR[0]="$train1"
ICV_ARR[1]="$train2"
ICV_ARR[2]="$train3"
ICV_ARR[3]="$train4"
ICV_ARR[4]="$tune1"
ICV_ARR[5]="$tune2"
ICV_ARR[6]="$tune3"
ICV_ARR[7]="$tune4"
ICV_ARR[8]="$tune5"
let "icvtrain1=${ICV_ARR[($j+1) % 9]}"
let "icvtrain2=${ICV_ARR[($j+2) % 9]}"
let "icvtrain3=${ICV_ARR[($j+3) % 9]}"
let "icvtrain4=${ICV_ARR[($j+4) % 9]}"
let "icvtune1=${ICV_ARR[($j+5) % 9]}"
let "icvtune2=${ICV_ARR[($j+6) % 9]}"
let "icvtune3=${ICV_ARR[($j+7) % 9]}"
let "icvtune4=${ICV_ARR[($j+8) % 9]}"
let "icvtest=${ICV_ARR[($j) % 9]}"

# echo "$FOLD $SUB_FOLD $SAYL_THRESHOLD $MINPOS $SEARCH_MODE $icvtrain1 $icvtrain2 $icvtrain3 $icvtrain4 $icvtune1 $icvtune2 $icvtune3 $icvtune4 $icvtest"

cat "$FS0"/"$EX""$icvtrain1".f \
  "$FS0"/"$EX""$icvtrain2".f \
  "$FS0"/"$EX""$icvtrain3".f \
  "$FS0"/"$EX""$icvtrain4".f > "$IPREFIX"train0.f
cat "$FS0"/"$EX""$icvtune1".f \
  "$FS0"/"$EX""$icvtune2".f \
  "$FS0"/"$EX""$icvtune3".f \
  "$FS0"/"$EX""$icvtune4".f > "$IPREFIX"tune0.f
cat "$FS0"/"$EX""$icvtest".f > "$IPREFIX"test0.f
cat "$IPREFIX"train0.f \
  "$FS1"/"$EX""$icvtrain1".f \
  "$FS1"/"$EX""$icvtrain2".f \
  "$FS1"/"$EX""$icvtrain3".f \
  "$FS1"/"$EX""$icvtrain4".f > "$IPREFIX"train.f
cat "$IPREFIX"tune0.f \
  "$FS1"/"$EX""$icvtune1".f \
  "$FS1"/"$EX""$icvtune2".f \
  "$FS1"/"$EX""$icvtune3".f \
  "$FS1"/"$EX""$icvtune4".f > "$IPREFIX"tune.f
cat "$IPREFIX"test0.f \
  "$FS1"/"$EX""$icvtest".f > "$IPREFIX"test.f

cat "$FS0"/"$EX""$icvtrain1".n \
  "$FS0"/"$EX""$icvtrain2".n \
  "$FS0"/"$EX""$icvtrain3".n \
  "$FS0"/"$EX""$icvtrain4".n > "$IPREFIX"train0.n
cat "$FS0"/"$EX""$icvtune1".n \
  "$FS0"/"$EX""$icvtune2".n \
  "$FS0"/"$EX""$icvtune3".n \
  "$FS0"/"$EX""$icvtune4".n > "$IPREFIX"tune0.n
cat "$FS0"/"$EX""$icvtest".n > "$IPREFIX"test0.n
cat "$IPREFIX"train0.n \
  "$FS1"/"$EX""$icvtrain1".n \
  "$FS1"/"$EX""$icvtrain2".n \
  "$FS1"/"$EX""$icvtrain3".n \
  "$FS1"/"$EX""$icvtrain4".n > "$IPREFIX"train.n
cat "$IPREFIX"tune0.n \
  "$FS1"/"$EX""$icvtune1".n \
  "$FS1"/"$EX""$icvtune2".n \
  "$FS1"/"$EX""$icvtune3".n \
  "$FS1"/"$EX""$icvtune4".n > "$IPREFIX"tune.n
cat "$IPREFIX"test0.n \
  "$FS1"/"$EX""$icvtest".n > "$IPREFIX"test.n
cat "$IPREFIX"train.f "$IPREFIX"tune.f > "$BK".f
cat "$IPREFIX"train.n  "$IPREFIX"tune.n > "$BK".n
#cat <<EOF
"$YAP" -l "$HOME"/ILP/sayu/aleph.yap  << EOF > "$PREFIX"/LOG.$i 2>&1
%yap_flag(gc_trace,verbose).
% saying which subset should be first thing,
read_all('$BK').
set(recordfile,'$PREFIX/$BG$i.trace').
vsc_set(sayu_dump_file('$PROB_FILE')).
vsc_set(sayu_min_recall_in_area($MIN_THRESHOLD)).
vsc_set(sayu_classifier_type($CLASSIFIER_TYPE)).
vsc_set(sayu_scoring_method($SCORING_METHOD)).
%vsc_set(sayu_start_seeds(200)).
vsc_set(sayu_prefix('$IPREFIX')).
set(nodes,$NODES).
set(verbosity,0).
set(minpos,$MINPOS).
vsc_set(sayu_mode($SEARCH_MODE)).
vsc_set(max_step(20)).
assert(user:sayl_threshold($SAYL_THRESHOLD)).
sayu_compute_strata0_sizes($IPREFIX).
Z is 3600, % 1 hour time limit
statistics(runtime,[T,_]), induce_sayu_timed(Z), statistics(runtime,[Tf,_]), D is Tf-T, format('ILP run took ~d msec.~n',[D]).


EOF
rm -f "$BK".[bfn] "$IPREFIX"train.[fn] "$IPREFIX"tune.[fn] "$IPREFIX"test.[fn] "$IPREFIX"train0.[fn] "$IPREFIX"tune0.[fn] "$IPREFIX"test0.[fn]
