package sayu;

public class CPTByte implements ProbDistribution{


    private int[] parentPos;
    private int pos;
    private int[] decode;

    private double[][] table;
    private double[][] logTable;
    private double[] totals;

    private double initialStrength = 5;

    private double[] pureProb;
    private double[] logPure;
    private double total;

    private boolean debug = false;

    public void setDebug(boolean debug){
	this.debug = debug;
    }

    public void display(){
	System.out.println("CPT for Position " + pos);
	if (parentPos.length == 0){
	    System.out.print("Pure prob: ");
	    for (int ix = 0; ix < pureProb.length; ix++){
		System.out.println(ix + ": " + Utils.round(pureProb[ix], 5));
	    }
	}
	else{
	    System.out.print("Parents: ");
	    for(int ix = 0; ix < parentPos.length; ix++){
		System.out.print(ix + ":" + parentPos[ix] + " ");
	    }
	    System.out.println();
	for(int ix = 0; ix < decode.length; ix++){
	    System.out.print(ix + ":" + decode[ix] + " ");
	}
	System.out.println();
	    for(int ix = 0; ix < table.length; ix++){
		for(int jx = 0; jx < table.length; jx++){
		    System.out.println(ix + " " + jx + " " + table[ix][jx] + " " + Utils.round(logTable[ix][jx], 5));
		}
	    }
	    System.out.println();
	}
    }

    public CPTByte(int thisPosition, int[] parentPositions,
		   Attribute[] attribs, byte[][] data){
	pos = thisPosition;
	parentPos = new int[parentPositions.length];
        System.arraycopy(parentPositions, 0, parentPos, 0, parentPos.length);
	int numvals = attribs[pos].getNumValues();
        int count;
        decode = new int[parentPositions.length];
        /*
	  Count the number of possible combinations for all of the attributes
	  that are parents of this node. 
         */

        if ( parentPos.length > 0){
            count = attribs[parentPositions[0]].getNumValues();
            decode[0] = 1;
            for(int ix = 1; ix < parentPositions.length; ix++){
                decode[ix] = count;
                count = count * attribs[parentPositions[ix]].getNumValues();

            }
	    table = new double[numvals][count];
	    logTable = new double[numvals][count];
            totals = new double[count];
	    int tableEntries = numvals * count;
            //initialize the probabilities in the table
            for(int ix = 0; ix < count; ix++){
                for(int jx = 0; jx < numvals; jx++){
                    table[jx][ix] = initialStrength / tableEntries;
                    totals[ix] += table[jx][ix];
                }
            }
            buildCPT(data);
	}
	else{
	    pureProb = new double[numvals];
	    logPure = new double[numvals];
            table = new double[numvals][];
            for(int ix = 0; ix < pureProb.length; ix++){
                pureProb[ix] = initialStrength / numvals;
                total += pureProb[ix];
            }
            buildPureProb(data);
 	}
    } 

    public CPTByte(int thisPosition, int[] parentPositions,
		   Attribute[] attribs, double[][] condProb){
	if (parentPositions.length != 1){
	    System.err.println("Parent Length is Incorrect");
	}
	pos = thisPosition;
	parentPos = new int[parentPositions.length];
        System.arraycopy(parentPositions, 0, parentPos, 0, parentPos.length);
	int numvals = attribs[pos].getNumValues();
      
        decode = new int[parentPositions.length];
	decode[0] = 1;
        
	int count = attribs[parentPositions[0]].getNumValues();
	table = new double[numvals][count];
	logTable = new double[numvals][count];
	totals = new double[count];

	//initialize the probabilities in the table
	for(int ix = 0; ix < count; ix++){
	    for(int jx = 0; jx < numvals; jx++){
		table[jx][ix] = condProb[jx][ix];
		totals[ix] += table[jx][ix];
	    }
	}
    } 

    public CPTByte(int thisPosition, int[] parentPositions,
		   Attribute[] attribs, Edge edge){
	pos = thisPosition;
	parentPos = new int[parentPositions.length];
        System.arraycopy(parentPositions, 0, parentPos, 0, parentPos.length);
	int numvals = attribs[pos].getNumValues();
        int count;
        decode = new int[parentPositions.length];
        /*
	  Count the number of possible combinations for all of the attributes
	  that are parents of this node. 
         */

        if ( parentPos.length == 0){
	    System.out.println("ERROR ");
	}
	count = attribs[parentPositions[0]].getNumValues();
	decode[0] = 1;
	for(int ix = 1; ix < parentPositions.length; ix++){
	    decode[ix] = count;
	    count = count * attribs[parentPositions[ix]].getNumValues();
	    
	}
	table = new double[numvals][count];
	logTable = new double[numvals][count];
	totals = new double[count];
	int tableEntries = numvals * count;
	//initialize the probabilities in the table
	double[][][] cnts = edge.getProbTable();
	int[] fixed = new int[2];
	for(int ix = 0; ix < table.length; ix++){
	    for(int jx = 0; jx < table[ix].length; jx++){
		table[ix][jx] = 0;
	    }
	}
	
	for(int ix = 0; ix < cnts.length; ix++){
	    for(int jx = 0; jx < cnts[ix].length; jx++){
		for(int kx = 0; kx < cnts[ix][jx].length; kx++){
		    if (edge.getStart() == pos){
			int index = jx * decode[0] + kx * decode[1];
			table[ix][index] += cnts[ix][jx][kx];
			totals[index] += table[ix][index];			
		    }
		    else{
			int index = ix * decode[0] + kx * decode[1];
			table[jx][index] += cnts[ix][jx][kx];
			totals[index] += table[jx][index];
		    }
		}
	    }
	}
	for(int ix = 0; ix < table.length; ix++){
	    for(int jx = 0; jx < table[ix].length; jx++){
		logTable[ix][jx] = Math.log(table[ix][jx] / totals[jx]);
	    }
	}

    }//end


   /**
       This method builds the index into the conditional probability table
       for the attribute at this pos. It takes into consideration the
       all the values of the parents

       @param example
       @param attributes
       @return the index into the CPT
     */
    private int buildIndex(byte[] example, int[] fixedValue){
			   
        int index = 0; //total index into table
         /*
	  go through each attribute that is a point of this node in the
	  graph and find its value.  the index into a table is found in 
	  this manner:
	  attribIndex * number of prior combinations of attribs 
         */
        int bound = parentPos.length;
        for(int ix = 0; ix < bound; ix++){
            if (fixedValue != null && fixedValue[ix] >= 0){
		index = index + (fixedValue[ix] * decode[ix]);
	    }
	    else{
		index = index + (example[parentPos[ix]] * decode[ix]);
	    }
        }
        return index;
    }//end of buildIndex


  /**
       This method builds the index into the conditional probability table
       for the attribute at this pos. It takes into consideration the
       all the values of the parents

       @param example
       @param attributes
       @return the index into the CPT
     */
    private int buildIndex(byte[] example, int pos, int fixedValue){
			   
        int index = 0; //total index into table
         /*
	  go through each attribute that is a point of this node in the
	  graph and find its value.  the index into a table is found in 
	  this manner:
	  attribIndex * number of prior combinations of attribs 
         */
        int bound = parentPos.length;
        for(int ix = 0; ix < bound; ix++){
            if (parentPos[ix] == pos){
		index = index + (fixedValue * decode[ix]);
	    }
	    else{
		index = index + (example[parentPos[ix]] * decode[ix]);
	    }
        }
        return index;
    }//end of buildIndex


   /**
       @param data is the train set used to build the counts

       note - the method doesn't actually compute the probilities, it just
       counts the occurances for each parent-child combination in the 
       training set

     */
    private void buildCPT(byte[][] data){
        int index; //the index for all of the parents
        /**
	   count all of the occurance for each pair of
         */ 
        for(int ix = 0; ix < data.length; ix++){
            /*
	      finds the values of all of the parents and the 
	      corresponding index into the conditional prob table
             */
            index = buildIndex(data[ix], null);
	    table[data[ix][pos]][index]++;
            totals[index]++; 
	}//end of for loop that counts occurances
	for(int ix = 0; ix < table.length; ix++){
	    for(int jx = 0; jx < table[ix].length; jx++){
		logTable[ix][jx] = Math.log(table[ix][jx] / totals[jx]);
	    }
	}
    }//end of method buildCPT

    /**
       @param train - the data to build up the probabilty
 
       this method builds a probability table for an attribute. it is a pure
       probability.  just finds the probability for each attribute value in 
       the training set.  

       note - the method doesn't actually compute the probility, it just
       counts the number of times each attribute value occurs in the 
       train set.
     */
    private void buildPureProb(byte[][] data){
        //build up a count for the occurance of each attribute value
        for(int ix = 0; ix < data.length; ix++){
            pureProb[data[ix][pos]]++;
        }//end of for loop for the count
        total += data.length;
	for(int ix = 0; ix < pureProb.length; ix++){
	    logPure[ix] = Math.log(pureProb[ix] / total);
	}
    }//end of method pureProb


    /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param attributes - the attributes for the feature vector
       @return the probility/condprob for the attribute at pos based on
               the values in the feature vector example

     */
    public final double getProb(byte[] example, int[] values, int fixThisValue){
	int thisValue = example[pos];
        if (fixThisValue >= 0 ){
	    thisValue = fixThisValue ;
        }
        //return the pure probability
        if (parentPos.length == 0){
            return (((double)pureProb[thisValue])/total);
        }
        //return the conditional probability
        else{
            int index = buildIndex(example, values);
	    return (((double)table[thisValue][index])/totals[index]);
        }

    }//getProb(DataElement example, Datapairs attributes[])
   /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param attributes - the attributes for the feature vector
       @return the probility/condprob for the attribute at pos based on
               the values in the feature vector example

     */
    public final double getProb(byte[] example, int position, int value){
	int thisValue = example[pos];

        //return the pure probability
        if (parentPos.length == 0){
            return (((double)pureProb[thisValue])/total);
        }
        //return the conditional probability
        else{
            int index = buildIndex(example, position, value);
	    return (((double)table[thisValue][index])/totals[index]);
        }

    }//getProb(DataElement example, Datapairs attributes[])


    /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param attributes - the attributes for the feature vector
       @return the probility/condprob for the attribute at pos based on
               the values in the feature vector example

     */
    public final double getLogProb(byte[] example, int[] values, int fixThisValue){
	int thisValue = example[pos];
        if (fixThisValue >= 0 ){
	    thisValue = fixThisValue ;
        }
        //return the pure probability
        if (parentPos.length == 0){
            return logPure[thisValue];
        }
        //return the conditional probability
        else{
            int index = buildIndex(example, values);
	    return logTable[thisValue][index];
        }

    }//getProb(DataElement example, Datapairs attributes[])


    /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param attributes - the attributes for the feature vector
       @return the probility/condprob for the attribute at pos based on
               the values in the feature vector example

     */
    public final double getLogProb(byte[] example, int position, int value){
	int thisValue = example[pos];
       
	//System.out.println("pos = " + pos + " ");
        //return the pure probability
        if (parentPos.length == 0){
            return logPure[thisValue];
        }
        //return the conditional probability
        else{
            int index = buildIndex(example, position, value);
	    return logTable[thisValue][index];
        }

    }//getProb(DataElement example, Datapairs attributes[])


    public int getFreeParameters(){
	System.out.println("NOT IMPLEMENTED");
	return 0;
    }

    public int getNumParents(){
	return parentPos.length;
    }

    public boolean parentOf(int position){
	boolean found = false;
	for(int ix = 0; ix < parentPos.length && !found; ix++){
	    if (parentPos[ix] == position){
		found = true;
	    }
	}
	return found;
    }

    public boolean changeParentPosition(int oldPosition, int newPosition){
	boolean found = false;
	for(int ix = 0; ix < parentPos.length && !found; ix++){
	    if (parentPos[ix] == oldPosition){
		found = true;
		parentPos[ix] = newPosition;
	    }
	}
	return found;
    }

    public void setThisPos(int position){
	this.pos = position;
    }
}