package sayu;
import java.util.*;
public class CVSayu{

    /*
      The models to perform cross validation on
    */
    private ILPSearch[] models;

    //final model with all the data
    private ILPSearch finalModel;

    //Score for computing a t-test
    private double[] oldScores;
    private double[] currentScores;


    //

    private int[] foldNumberPos;
    private int[] foldNumberNeg;
    private int testPos = 0;
    private int testNeg = 0;
    private int[] posCounts;
    private int[] negCounts;


    /**
       @param modelType (NB, Tan, FullTan)
       @param numFolds the number of folds to create
       @param trainPosFoldNumber which fold each positive training example should be assigned to
       @param trainNegFoldNumber which fold each negative training example should be assigned to
       @param testPos the number of positive examples in the test set
       @param testNeg the number of negative examples in the test set
    */
    public CVSayu(int modelType, int numFolds, int[] trainPosFoldNum, int[] trainNegFoldNum, int testPos, int testNeg){

	try {
	    models = new ILPSearch[numFolds];
	    oldScores = new double[numFolds];
	    currentScores = new double[numFolds];
	    posCounts = new int[numFolds];
	    negCounts = new int[numFolds];
	    for(int ix = 0; ix < posCounts.length; ix++){
		posCounts[ix] = 0;
		negCounts[ix] = 0;
	    }
	    for(int ix = 0; ix < trainPosFoldNum.length; ix++){
		posCounts[trainPosFoldNum[ix]]++;
	    }
	    for(int ix = 0; ix < trainNegFoldNum.length; ix++){
		negCounts[trainNegFoldNum[ix]]++;
	    }
	    this.foldNumberPos = trainPosFoldNum;
	    this.foldNumberNeg = trainNegFoldNum;
	    this.testPos = testPos;
	    this.testNeg = testNeg;
	    for(int ix = 0; ix < models.length; ix++){
		if (modelType == Constants.NB){
		    models[ix] = new IncNB(trainPosFoldNum.length-posCounts[ix], trainNegFoldNum.length-negCounts[ix], posCounts[ix], negCounts[ix], testPos,testNeg);
		}
		else if (modelType == Constants.TAN){
		    models[ix] = new IncTan(trainPosFoldNum.length-posCounts[ix], trainNegFoldNum.length-negCounts[ix], posCounts[ix], negCounts[ix], testPos,testNeg);
		}
	    }
	    if (modelType == Constants.NB){
		finalModel = new IncNB(trainPosFoldNum.length, trainNegFoldNum.length, trainPosFoldNum.length, trainNegFoldNum.length, testPos,testNeg);
		
	    }
	    else if (modelType == Constants.TAN){
		finalModel = new IncTan(trainPosFoldNum.length, trainNegFoldNum.length, trainPosFoldNum.length, trainNegFoldNum.length, testPos,testNeg);
	    
	    }
        } catch (Exception e) {
            e.printStackTrace();
        }
       

    }






    /**
       This method introduces a new feature into the train set for
       each fold.
       @param trainPosVector the new feature for the positive training examples
       @param trainNegVector the new feature for the negative training examples
       @param testPosVector the new feature for the positive test examples
       @param testNegVector the new feature for the negative test examples
       @param scoreFunction
       @return the score for each fold according to 'scoreFunction'
    */
    public double[] addAttribute(byte[] trainPosVector, byte[] trainNegVector, byte[] testPosVector, byte[] testNegVector, int scoreFunction, int numValues){
	try {
	    double[] scores = new double[models.length];
	    byte[][][] dividedPos = split(trainPosVector, foldNumberPos, posCounts, foldNumberPos.length);
	    byte[][][] dividedNeg = split(trainNegVector, foldNumberNeg, negCounts, foldNumberNeg.length);

	    for(int ix = 0; ix < models.length; ix++){
		scores[ix] = models[ix].addAttribute(dividedPos[ix][0], dividedNeg[ix][0], dividedPos[ix][1], dividedNeg[ix][1], testPosVector, testNegVector, scoreFunction, numValues);
		currentScores[ix] = scores[ix];
	    
	    }
	    finalModel.addAttribute(trainPosVector, trainNegVector, trainPosVector, trainNegVector, testPosVector, testNegVector, scoreFunction, numValues);
	    return scores;
        } catch (Exception e) {
            e.printStackTrace();
	    return null;
        }
	/*
	  Sayu_maxMIR.java:
	  double Bagrsq(double[] pred, Double[] true, double[] mean)
	*/
	//return Sayu_maxMIR.Bagrsq(newPredictions, trueOutput, index, means);
    } 

   /**
    
       This method permanently adds the most recently introduced
       feature into feature set.  Notifies the learner
       of the change.
    */
     public void commit(){
	System.arraycopy(currentScores, 0, oldScores, 0, currentScores.length);
	for(int ix = 0; ix < models.length; ix++){
	    models[ix].commit();
	}
	finalModel.commit();
    }


    
    public double getScore(int dataSet, int scoreFunction){
	return finalModel.getScore(dataSet,scoreFunction);
    }
    public double getCurrentScore(int dataSet, int scoreFunction){
	//return getScore(dataSet,scoreFunction,currentModel);
	return -1;
    }
 
    /**
       @param dataSet whether to return the probabilities for the train, tune or test set.
       @return the probability that each example in the desired dataSet is of the positive class. 
     */
    public  double[] getProbs(int dataSet){
        try{
            return finalModel.getProbs(dataSet);
        }
        catch(Exception exc){
            exc.printStackTrace();
        }
        return null;
    }//end of method getProbs(int dataSet)



    public  void rulesEval(){
	/*
        System.out.println("scored " + 
			   finalModel.getNumFeatsScored() + 
			   " rules");
	*/
    }

    public  int numFeatsScored(){
        //return finalModel.getNumFeatsScored();
	return -1;
    }

    public  void terminate(String fileStem){
        try{
            System.out.println("calling terminate: Not Implemented");
	    /*
	    for(int ix = 0; ix < models.length; ix++){
		models[ix].terminate(fileStem);
	    }
            finalModel.terminate(fileStem);
	    */
        }
        catch(Exception exc){
            exc.printStackTrace();
        }
    }

    /*
    public  double[] disjoin(int scoreFunction){
        try{
            return search.disjoin(scoreFunction);
        }
        catch(Exception exc){
            exc.printStackTrace();
        }
        return new double[3];
    }
    */

    public  void newHead(int type){
	for(int ix = 0; ix < models.length; ix++){
	    models[ix].newHead((byte)type);
	}
	
        finalModel.newHead((byte)type);
    }

    public  double[] removeFeature(int feature, int scoreFunction){
	double[] scores = new double[models.length];
        try {
	    for(int ix = 0; ix < models.length; ix++){
		scores[ix] = models[ix].removeAttribute(feature,scoreFunction);
	    }      
	    finalModel.removeAttribute(feature,scoreFunction);
        } 
	catch (Exception e) {
            e.printStackTrace();
        }
        return scores;
    }

    public double[] replaceAttribute(byte[] trainPosVector, byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector, byte[] testNegVector, int replacePos, int scoreFunction, int numValues){
        
        try {
	    double[] scores = new double[models.length];
	    byte[][][] dividedPos = split(trainPosVector, foldNumberPos, posCounts, foldNumberPos.length);
	    byte[][][] dividedNeg = split(trainNegVector, foldNumberNeg, negCounts, foldNumberNeg.length);

	    for(int ix = 0; ix < models.length; ix++){
		scores[ix] = models[ix].replaceAttribute(dividedPos[ix][0], dividedNeg[ix][0], dividedPos[ix][1], dividedNeg[ix][1], testPosVector, testNegVector, replacePos, scoreFunction, numValues);
		currentScores[ix] = scores[ix];
	    
	    }
	    finalModel.replaceAttribute(trainPosVector, trainNegVector, tunePosVector, tuneNegVector, testPosVector, testNegVector, replacePos, scoreFunction, numValues);
	    return scores;
        } 
	catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }



    /**
       divides a feature vector to create one feature vector each fodl
    */
    private byte[][][] split(byte[] train, int[] foldNumber, int[] counts, int total){
	byte[][][] data = new byte[models.length][2][];
	for(int ix = 0; ix < data.length; ix++){
	    int trainPos =0;
	    int tunePos = 0;
	    data[ix][0] = new byte[ total - counts[ix]];
	    data[ix][1] = new byte[counts[ix]];
	    for(int jx = 0; jx < foldNumber.length; jx++){
		if (ix == foldNumber[jx]){
		    data[ix][1][tunePos] = train[jx];
		    tunePos++;
		} 
		else{
		    data[ix][0][trainPos] = train[jx];
		    trainPos++;
		}
	    }
	}
	return data;
    }


   /**
       divides a feature vector to create one feature vector each fodl
    */
    private int[][][][] createTrain(int[][] pos, int[][] neg, int[] foldNumberPos, int[] foldNumberNeg, int[] countPos, int[] countNeg){
	int totalPos = foldNumberPos.length;
	int totalNeg = foldNumberNeg.length;
	int fvLength = neg[0].length;
	int[][][][] data = new int[models.length][][][];
	for(int ix = 0; ix < data.length; ix++){
	    int trainPos =0;
	    int tunePos = 0;
	    int tuneLen = countPos[ix]+ countNeg[ix];
	    int trainLen = totalPos + totalNeg - tuneLen;
	    data[ix][0] = new int[trainLen][fvLength];
	    data[ix][1] = new int[tuneLen][fvLength];
	    for(int jx = 0; jx < foldNumberNeg.length; jx++){
		for(int kx = 0; kx < neg[jx].length; kx++){
		    if (ix == foldNumberNeg[jx]){
			data[ix][1][tunePos][kx] = neg[jx][kx];
			tunePos++;
		    } 
		    else{
			data[ix][0][trainPos][kx] = neg[jx][kx];
			trainPos++;
		    }
		}
	    }
	    for(int jx = 0; jx < foldNumberPos.length; jx++){
		for(int kx = 0; kx < neg[jx].length; kx++){
		    
		    if (ix == foldNumberPos[jx]){
			data[ix][1][tunePos][kx] = pos[jx][kx];
			tunePos++;
		    } 
		    else{
			data[ix][0][trainPos][kx] = pos[jx][kx];
			trainPos++;
		    }
		}
	    }
	}
	return data;
    }



}
