/**
   @author Jesse Davis
   (c) 5/2003

 */
package sayu;
import java.io.*;
public class CondProb{

    public static boolean cDebug = false;

    //holds the indices of the parent attributes in the feature vector
    private int[] parentPos;

    //holds the index of this attribute in the feature vector
    private int thisPos;

    //holds the index of the classification.  the classfication needs to be
    //access differently
    private int classPos;

    /**
       holds on to how to index into the conditional probability table
       for each parent attribute
     */
    private int[] decode;
    /**
       holds the occurance for each pair of possible parent - child combinations
     */
    private double[][] table;

    /**
       total for each row in the CPT
     */
    private double[] totals;


    /*
      occurances for each attribute Value, if this attribute has no parents
      in the bayes net
     */
    private double[] pureProb;

    //total number of examples in the train set
    private double total;
    //private double total;

    private double init_count = 5;


    private double[][] logProbs;

    public CondProb(int thisPosition, int[] parentPositions,  int classPos, 
            int[][] set){

    }
    public CondProb(int thisPosition, int[] parentPositions,  int classPos, 
            int[][] set, Attribute[] attribs){

        this.classPos = classPos;
        this.thisPos = thisPosition;
	if (attribs[thisPos] == null){
	    System.out.print("Attribs @ position " + thisPos + " is null ");
	    System.out.print("class position = " + classPos);
	    System.out.println(" attributes.length = " + attribs.length);
	}
        int numvals = attribs[thisPos].getNumValues();

        parentPos = new int[parentPositions.length];
        System.arraycopy(parentPositions, 0, parentPos, 0, parentPos.length);

        int count;
        decode = new int[parentPositions.length];
        /*
	  Count the number of possible combinations for all of the attributes
	  that are parents of this node. 
         */

        if ( parentPos.length > 0){
            count = attribs[parentPositions[0]].getNumValues();
            decode[0] = 1;
            for(int ix = 1; ix < parentPositions.length; ix++){
                decode[ix] = count;
                count = count * attribs[parentPositions[ix]].getNumValues();

            }
            table = new double[numvals][count];
            totals = new double[count];
	    double div = numvals * count;
            //initialize the probabilities in the table
            //RED FLAG - M estimates??
            for(int ix = 0; ix < count; ix++){
                for(int jx = 0; jx < numvals; jx++){
                    table[jx][ix] = init_count / div;
                    totals[ix] += table[jx][ix];
                }
            }
            buildCPT(set);
            //logProbs


        }
        /*
	  This is a pure probability. no parents, so just count the occurances
	  of each value.
         */
        else{
            pureProb = new double[numvals];
            table = new double[numvals][];
	    
            for(int ix = 0; ix < pureProb.length; ix++){
                pureProb[ix] = init_count / (double)numvals;
                total += pureProb[ix];
            }
            buildPureProb(set);


        }
        buildLogProb();
    }

    public CondProb(int thisPosition, int[] parentPositions,  int classPos, 
            byte[][] set){

    }
    public CondProb(int thisPosition, int[] parentPositions,  int classPos, 
            byte[][] set, Attribute[] attribs){

        this.classPos = classPos;
        this.thisPos = thisPosition;
        int numvals = attribs[thisPos].getNumValues();

        parentPos = new int[parentPositions.length];
        System.arraycopy(parentPositions, 0, parentPos, 0, parentPos.length);

        int count;
        decode = new int[parentPositions.length];
        /* Count the number of possible combinations for all of the attributes
	     * that are parents of this node. 
         */
        //System.err.println("Called CondProb "+thisPosition);
        if ( parentPos.length > 0 && attribs[parentPositions[0]] != null){ //EL
            //System.err.println("CondProb "+parentPositions[0]+" "+attribs[parentPositions[0]]+" "+attribs.length);
            count = attribs[parentPositions[0]].getNumValues();
            decode[0] = 1;
            for(int ix = 1; ix < parentPositions.length; ix++){
                decode[ix] = count;
		if (parentPositions == null){
		    System.out.println("Par Pos is null");
		}
		if (attribs[parentPositions[ix]] == null){
		    if (attribs == null){
			System.out.println("attribs array is null");
		    }
		    System.out.println("attribs is null at " + parentPositions[ix]);
		}
                count = count * attribs[parentPositions[ix]].getNumValues();

            }
            table = new double[numvals][count];
            totals = new double[count];
	    double div = numvals * count;
            //initialize the probabilities in the table
            //RED FLAG - M estimates??
            for(int ix = 0; ix < count; ix++){
                for(int jx = 0; jx < numvals; jx++){
                    table[jx][ix] = init_count /div;
                    totals[ix] += table[jx][ix];
                }
            }
            buildCPT(set);

        }
        /*
	  This is a pure probability. no parents, so just count the occurances
	  of each value.
         */
        else{
            pureProb = new double[numvals];
            table = new double[numvals][];
            for(int ix = 0; ix < pureProb.length; ix++){
                pureProb[ix] = init_count / (double)numvals;
                total += pureProb[ix];
            }
            buildPureProb(set);
        }
        buildLogProb();
    }

    /**
       This method builds the index into the conditional probability table
       for the attribute at this thisPos. It takes into consideration the
       all the values of the parents

       @param example
       @param attributes
       @return the index into the CPT
     */
    private int buildIndex(boolean useClass,  byte[] vals,
            int out){
        int position; //the current parent being evaluated
        int index = 0; //total index into table
        int featVal = 0; //index component for one attribute
        int tempIndex =0;
        /*
	  go through each attribute that is a point of this node in the
	  graph and find its value.  the index into a table is found in 
	  this manner:
	  attribIndex * number of prior combinations of attribs 
         */
        int bound = parentPos.length;
        for(int ix = 0; ix < bound; ix++){


            position = parentPos[ix];
            featVal = vals[position];
            if ((position == classPos) && useClass){
                featVal = out;
            }
            if (cDebug && thisPos == 36){
                System.out.println("position: " + position + " feat val " + featVal);
            }
            //build up the index
            tempIndex = featVal * decode[ix];
            index += tempIndex;
        }
        return index;
    }//end of buildIndex


    /**
       This method builds the index into the conditional probability table
       for the attribute at this thisPos. It takes into consideration the
       all the values of the parents

       @param example
       @param attributes
       @return the index into the CPT
     */
    private int buildIndex(boolean useClass,  int[] vals,
            int out){
        int position; //the current parent being evaluated
        int index = 0; //total index into table
        int featVal = 0; //index component for one attribute
        int tempIndex =0;
        /*
	  go through each attribute that is a point of this node in the
	  graph and find its value.  the index into a table is found in 
	  this manner:
	  attribIndex * number of prior combinations of attribs 
         */
        int bound = parentPos.length;
        for(int ix = 0; ix < bound; ix++){
            position = parentPos[ix];
            featVal = vals[position];
            if ((position == classPos) && useClass){
                featVal = out;
            }
            //build up the index
            tempIndex = featVal * decode[ix];
            index += tempIndex;
        }
        return index;
    }//end of buildIndex

    /**

       @return
     */
    private int[] unbuildIndex(int index){
        int[] vals = new int[decode.length];
        int remainder;
        int quo = index;
        for(int ix = (decode.length  - 1); ix >=0; ix--){
            remainder = quo % decode[ix];
            quo = quo / decode[ix];
            vals[ix] = quo;
            quo = remainder;
        }
        return vals;
    }//end unbuildIndex();


    /**
       @param train - the train set to build the cpt on
       @param attributes - the list of attributes for the feature vectors


       note - the method doesn't actually compute the probilities, it just
       counts the occurances for each parent-child combination in the 
       training set

     */
    private void buildCPT(byte[][] data){
        int index; //the index for all of the parents
        int thisIndex; //the index for the value of the attribute at thisPos 

        /**
	   count all of the occurance for each pair of
         */ 
        int bound = data.length;
        for(int ix = 0; ix < bound; ix++){
            /*
	      finds the values of all of the parents and the 
	      corresponding index into the conditional prob table
             */
            index = buildIndex(false, data[ix], -1);
            thisIndex = data[ix][thisPos];
            if (index >= table[0].length){
                System.out.println("problem Index "+ table.length + " " + table[0].length);
            }
            if (thisIndex >= table.length){
                System.out.println("problem thisIndex " + table.length + " " + table[0].length + " " + thisPos);
            }
            table[thisIndex][index]++;
            totals[index]++; 

        }//end of for loop that counts occurances

    }//end of method buildCPT


    /**
       @param train - the train set to build the cpt on
       @param attributes - the list of attributes for the feature vectors


       note - the method doesn't actually compute the probilities, it just
       counts the occurances for each parent-child combination in the 
       training set

     */
    private void buildCPT(int[][] data){
        int index; //the index for all of the parents
        int thisIndex; //the index for the value of the attribute at thisPos 

        /**
	   count all of the occurance for each pair of
         */ 
        int bound = data.length;
        for(int ix = 0; ix < bound; ix++){
            /*
	      finds the values of all of the parents and the 
	      corresponding index into the conditional prob table
             */
            index = buildIndex(false, data[ix], -1);
            thisIndex = data[ix][thisPos];
            if (index >= table[0].length){
                System.out.println("problem Index "+ table.length + " " + table[0].length + " " + thisPos + " value " + thisIndex + " " + ix + " " + data[ix][17] + " " + data[ix][35]);
            }
            if (thisIndex >= table.length){
                System.out.println("problem thisIndex " + table.length + " " + table[0].length + " " + thisPos);
            }
            table[thisIndex][index]++;
            totals[index]++; 

        }//end of for loop that counts occurances

    }//end of method buildCPT

    /**
       @param train - the data to build up the probabilty
       @param attribute - the attribute for which table is being built

       this method builds a probability table for an attribute. it is a pure
       probability.  just finds the probability for each attribute value in 
       the training set.  

       note - the method doesn't actually compute the probility, it just
       counts the number of times each attribute value occurs in the 
       train set.
     */
    private void buildPureProb(byte[][] data){
        int bound = data.length;
        //build up a count for the occurance of each attribute value
        for(int ix = 0; ix < bound; ix++){
            pureProb[data[ix][thisPos]]++;
        }//end of for loop for the count
        total += data.length;

    }//end of method pureProb

    /**
       @param train - the data to build up the probabilty
       @param attribute - the attribute for which table is being built

       this method builds a probability table for an attribute. it is a pure
       probability.  just finds the probability for each attribute value in 
       the training set.  

       note - the method doesn't actually compute the probility, it just
       counts the number of times each attribute value occurs in the 
       train set.
     */
    private void buildPureProb(int[][] data){
        int bound = data.length;
        //build up a count for the occurance of each attribute value
        for(int ix = 0; ix < bound; ix++){
            if (data[ix][thisPos] >= pureProb.length){
                System.out.println(thisPos);
            }
            pureProb[data[ix][thisPos]]++;
        }//end of for loop for the count
        total += data.length;

    }//end of method pureProb

    private void buildLogProb(){
        if (parentPos.length == 0){
            logProbs = new double[1][pureProb.length];
            for(int ix = 0; ix < pureProb.length; ix++){
                logProbs[0][ix] = Math.log(((double)pureProb[ix] /total));
            }
        }
        else{
            logProbs = new double[table.length][table[0].length];
            for(int ix = 0; ix < table.length; ix++){
                for(int jx = 0; jx < table[ix].length; jx++){
                    logProbs[ix][jx] = Math.log(((double)table[ix][jx] /totals[jx]));
                }
            }
        }
    }

    /***********************/
    public double getLogProb(int[] ex,boolean useClass, int out){
        int thisIndex;//the index for the value of the attribute at thisPos 
        int index; //index into the CPT based on the parent value

        if (thisPos == classPos && useClass){

            thisIndex = out;
        }
        else{
            thisIndex = ex[thisPos];
        }
        if (cDebug){// && thisPos == 36){
            System.out.print("thisPos " + thisPos + " thisIndex: " + thisIndex);
        }
        //return the pure probability

        if (parentPos.length == 0){
            if (cDebug){// && thisPos == 36){
                System.out.println(" pureProb ");
            }
            return logProbs[0][thisIndex];
        }
        //return the conditional probability
        else{
            index = buildIndex(useClass,  ex, out);
            if (cDebug){// && thisPos == 36){
                System.out.println(" index: " + index);
            }

            return logProbs[thisIndex][index];
        }
    }

    /***********************/
    public double getLogProb(byte[] ex,boolean useClass, int out){
        int thisIndex;//the index for the value of the attribute at thisPos 
        int index; //index into the CPT based on the parent value

        if (thisPos == classPos && useClass){

            thisIndex = out;
        }
        else{
            thisIndex = ex[thisPos];
        }
        if (cDebug){// && thisPos == 36){
            System.out.print(" thisPos " + thisPos + " thisIndex: " + thisIndex);
        }

        //return the pure probability
        if (parentPos == null){
            System.out.println(thisPos + " parentPos is null");
        }
        if (parentPos.length == 0){
            return logProbs[0][thisIndex];
        }
        //return the conditional probability
        else{
            index = buildIndex(useClass,  ex, out);
            if (cDebug){// && thisPos == 36){
                System.out.println(" index: " + index);
            }
            return logProbs[thisIndex][index];
        }
    }

    public void setThisPos(int thisPos){
        this.thisPos = thisPos;
    }

    /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param attributes - the attributes for the feature vector
       @return the probility/condprob for the attribute at thisPos based on
               the values in the feature vector example

     */
    public final double getProb(boolean useClass, 
            int[] ex, int out){
        int thisIndex;//the index for the value of the attribute at thisPos 
        int index; //index into the CPT based on the parent value

        if (thisPos == classPos && useClass){

            thisIndex = out;
            if (thisIndex < 0){
                System.out.println("thisIndex " + out);
            }
        }
        else{

            thisIndex = ex[thisPos];
            if (thisIndex < 0){
                System.out.println(thisPos + " " + ex[thisPos]);
            }
        }


        //return the pure probability
        if (parentPos.length == 0){
            if ((pureProb[thisIndex] == 0) || (total == 0)){
                System.out.println("total is " + total + 
                        " pureprob cnt " + pureProb[thisIndex]);
            }
            return (((double)pureProb[thisIndex])/total);
        }
        //return the conditional probability
        else{
            index = buildIndex(useClass,  ex, out);
            if (index < 0){
                System.out.println("index less than zero");
            }
            if (thisIndex < 0){
                System.out.println("thisIndex is bad");
            }
            if (table[thisIndex][index] == 0 || totals[index] == 0){
                /*
		  System.out.println("why is an index zero?");
		  System.out.println("this pos is " + thisPos + 
		  " this index is " + index + 
		  " table count " + table[thisIndex][index]
		  + " totals " + totals[index]);
		  System.out.println("table[0].length = " + table[0].length);
                 */
                return 0;
            }
            else{
                return (((double)table[thisIndex][index])/totals[index]);
            }
        }

    }//getProb(DataElement example, Datapairs attributes[])


    /**
       Calculates the number of free Parameters in this CPT.  Assumes
       that all of the variables are binary.

       @return number of free parameters
     */
    public int numFreeParams(){
        if (parentPos.length == 0){
            return pureProb.length -1;
        }
        else{
            return table[0].length * (table.length-1);
        }
    }//end of method numFreeParams

    public int getNumParents(){
        return parentPos.length;
    }

    public int[] getParents(){
        return parentPos;
    }

    public double[] getPureProb(){
        return pureProb;
    }

    public double[][] getTable(){
        return table;
    }

    public boolean compare(CondProb cpt){
        double[][] table2 = cpt.getTable();
        double[] pp2 = cpt.getPureProb();
        if (cpt.getNumParents() == 0){
            for(int ix =0; ix < pp2.length; ix++){
                if (pp2[ix] != pureProb[ix]){
                    System.out.println("error " + thisPos);
                    return false;
                }
            }
        }
        else{
            for(int ix =0; ix < table2.length; ix++){
                for(int jx = 0; jx < table2[ix].length; jx++){
                    if (table[ix][jx] != table2[ix][jx]){
                        System.out.println("error " + thisPos);
                        return false;
                    }
                }
            }
        }
        return true;

    }
}//end of class CondProb

