package sayu;
import java.util.*;
import java.io.*;
public class Data{

    private boolean debug = false; 

    private int[][] trainInts;
    private int[][] tuneInts;
    private int[][] testInts;
    
    private Element[] trainE;
    private Element[] tuneE;
    private Element[] testE;

    private Datapairs[] dp;
    private int outputPosition;

    public Element[] getTrainElement(){
	return trainE;
    }

    public Element[] getTuneElement(){
	return tuneE;
    }

    public Element[] getTestElement(){
	return testE;
    }

    public int[][] getTrainInts(boolean why){
	return trainInts;
    }
    public int[][] getTestInts(){
	return testInts;
    }
    public int[][] getTuneInts(){
	return tuneInts;
    }

    public int[][] combineTrainTune(){
	int[][] n = new int[trainInts.length+tuneInts.length][];
	for(int ix = 0; ix < trainInts.length; ix++){
	    System.arraycopy(trainInts[ix], 0, n[ix], 0, trainInts[ix].length);
	}
	int off = trainInts.length;
	for(int ix = off; ix < n.length; ix++){
	    System.arraycopy(tuneInts[ix-off], 0, n[ix], 0, tuneInts[ix-off].length);
	}
	return n;
    }

    public int getOutputPosition(){
	return outputPosition;
    }
    public Datapairs[] getDP(){
	return dp;
    }

    public Data(String stem, String train, String tune, String test, String dataSuff, String keySuff,
		String names, String sep, boolean discrete, boolean isInt){
	nameRead(stem + names);
	LinkedList<Element> exs = new LinkedList<Element>();
	LinkedList<Element> tuneExs = new LinkedList<Element>();
	LinkedList<Element> testExs = new LinkedList<Element>();
	//Read in the data
	
	readDataKey(stem + test + dataSuff , stem + test + keySuff, testExs, sep);
	readDataKey(stem + train + dataSuff, stem + train + keySuff,exs, sep);
	readDataKey(stem + tune + dataSuff, stem + tune + keySuff, tuneExs, sep);
	trainE = new Element[exs.size()];
	tuneE = new Element[tuneExs.size()];
	testE = new Element[testExs.size()];
	trainInts = convertListInts(exs, trainE);
	tuneInts = convertListInts(tuneExs, tuneE);
	testInts = convertListInts(testExs, testE);	

    }
    /**
       @param suffix
       @param start
       @param stop
       @param test
       @param sep
       @param discrete
       @param isInt
    */
    public Data(String stem, String suffix, int[] train,
		int test, String names, String sep,
		boolean discrete, boolean isInt){
	
	nameRead(names);
	LinkedList<Element> exs = new LinkedList<Element>();
	LinkedList<Element> tuneExs = new LinkedList<Element>();
	LinkedList<Element> testExs = new LinkedList<Element>();
	//Read in the data
	if (train == null){
	    System.out.println("train null");
	}
	if (train.length == 0){
	    System.out.println("train length 0");
	}
	readDataKey(stem + test + suffix, stem + test + ".key", testExs, sep);
		 
	for(int ix = 0; ix < train.length; ix++){
	    if (debug){
		System.out.println(stem + ix + suffix);
		System.out.println(stem + ix + ".key");
	    }
	    if (train[ix] == 0){
		readDataKey(stem + ix + suffix, stem + ix + ".key",exs, sep);
	    }
	    else if (train[ix] == 1){
		readDataKey(stem + ix + suffix, stem + ix + ".key", tuneExs, sep);
	    }
	}
	trainE = new Element[exs.size()];
	tuneE = new Element[tuneExs.size()];
	testE = new Element[testExs.size()];
	trainInts = convertListInts(exs, trainE);
	tuneInts = convertListInts(tuneExs, tuneE);
	testInts = convertListInts(testExs, testE);	
    }

    public void simpleJoinAgg(String path, String data2){
	String sep = ",";
	try{
	    for(int ix = 0; ix < 10; ix++){
		for(int xx = 0; xx < 3; xx++){
		    LinkedList<Element> exs1 = new LinkedList<Element>();
		    LinkedList<Element> exs2 = new LinkedList<Element>();
		    RandomAccessFile dataOut;
		    RandomAccessFile keyOut;
		    RandomAccessFile newNames;
		    if (xx == 0){
			dataOut = new RandomAccessFile(path + "trainFull" + ix + ".data", "rw");
			keyOut = new RandomAccessFile(path +"trainFull" + ix + ".key", "rw");
			
			readDataKey(path + "trainAgg" + ix + ".data", path + "trainAgg" + ix +".key",exs1, sep);
			readDataKey(data2 + ix + "_bmap.train", data2 + ix + "_keys.train",exs2, sep);

		    }
		    else if (xx == 1){
			dataOut = new RandomAccessFile(path + "tuneFull" + ix + ".data", "rw");
			keyOut = new RandomAccessFile(path +"tuneFull" + ix + ".key", "rw");

			readDataKey(path + "tuneAgg" + ix + ".data", path + "tuneAgg" + ix +".key",exs1, sep);
			readDataKey(data2 + ix + "_bmap.tune1", data2+ ix + "_keys.tune1",exs2, sep);
			readDataKey(data2 + ix + "_bmap.tune2", data2 + ix + "_keys.tune2",exs2, sep);

		    }
		    else{
			dataOut = new RandomAccessFile(path + "testFull" + ix + ".data", "rw");
			keyOut = new RandomAccessFile(path +"testFull" + ix + ".key", "rw");

			readDataKey(path + "testAgg" + ix + ".data", path + "testAgg" + ix +".key",exs1, sep);
			readDataKey(data2 + ix + "_bmap.test", data2 + ix + "_keys.test",exs2, sep);
		    }
		    Element[] e1 = convertListElements(exs1);
		    Element[] e2 = convertListElements(exs2);
		    int numRules = 0;
		    for(int jx = 0; jx < e1.length; jx++){
			if (e1[jx].sameKey(e2[jx])){
			    if (e1[jx].getKey().equals("")){
				System.out.println("error - key");
			    }
			    keyOut.writeBytes(e1[jx].getKey() + "\n");
			    int[] fv = e1[jx].getFV();
			    if (jx == 0){
				System.out.println("num feats " + fv.length);
			    }
			    
			    for(int kx = 0; kx < fv.length-1; kx++){
				dataOut.writeBytes(fv[kx] +",");
			    }
			    fv = e2[jx].getFV();
			    for(int kx = 0; kx < fv.length-1; kx++){
				dataOut.writeBytes(fv[kx] +",");
			    }
			    
			    numRules = fv.length-1;
			    dataOut.writeBytes(fv[fv.length-1] + "\n");
			}
			else{
			    System.err.println("error " + ix + " " + jx);
			    System.err.println(e1[jx].getKey() + " " + e2[jx].getKey() + " " + e1[jx+1].getKey());
			    
			    System.exit(0);
			}
		    }

		    if (xx == 0){
			newNames = new RandomAccessFile(path +"fullFold" + ix + ".names", "rw");
			String nameO = path + "fold" + ix + ".names";
			RandomAccessFile names = new RandomAccessFile(nameO, "r");
			LinkedList<String> namesList = new LinkedList<String>();
			//System.out.println(dp.length);
			String nLine = names.readLine();
			while(nLine != null){
			    namesList.add(nLine);
			    nLine = names.readLine();
			}
			names.close();
			String[] namesA = new String[namesList.size()];
			for(int nx = 0; nx < namesA.length; nx++){
			    namesA[nx] = (String)namesList.removeFirst();
			}
			
			for(int nx = 0; nx < (namesA.length-1); nx++){
			    newNames.writeBytes(namesA[nx] + "\n");
			}
			for(int nx = 0; nx < numRules; nx++){
			    newNames.writeBytes("r" + nx + " discrete 0,1\n");
			}
			newNames.writeBytes(namesA[namesA.length-1]);
			newNames.close();
		    }
		
		    dataOut.close();
		    keyOut.close();
		    
		    


		}//end of xx
	    }//end of ix
	}
	catch(IOException exc){
	    System.out.println(exc.toString());
	    exc.printStackTrace();	    
	}
    }

    public void simpleJoin(String data1, String key1, String data2, String key2, String dataFile, String keyFile, String nameO, String nameN){
	String sep = ",";
	try{
	    LinkedList<String> namesList = new LinkedList<String>();
	    RandomAccessFile names = new RandomAccessFile(nameO, "r");
	    nameRead(nameO);
	    //System.out.println(dp.length);
	    String nLine = names.readLine();
	    while(nLine != null){
		namesList.add(nLine);
		nLine = names.readLine();
	    }
	    names.close();
	    String[] namesA = new String[namesList.size()];
	    for(int ix = 0; ix < namesA.length; ix++){
		namesA[ix] = namesList.removeFirst();
	    }

	    for(int ix = 0; ix < 10; ix++){
		LinkedList<Element> exs1 = new LinkedList<Element>();
		LinkedList<Element> exs2 = new LinkedList<Element>();
		RandomAccessFile dataOut = new RandomAccessFile(dataFile + ix + ".data", "rw");
		RandomAccessFile keyOut = new RandomAccessFile(keyFile + ix + ".key", "rw");
		RandomAccessFile newNames = new RandomAccessFile(nameN + ix + ".names", "rw");
		readDataKey(data1 + ix + ".data", key1 + ix + ".key",exs1, sep);
		readDataKey(data2 + ix + "_bmap.test", key2 + ix + "_keys.test",exs2, sep);
		Element[] e1 = convertListElements(exs1);
		Element[] e2 = convertListElements(exs2);
		int numRules = 0;
		for(int jx = 0; jx < e1.length; jx++){
		    if (e1[jx].sameKey(e2[jx])){
			if (e1[jx].getKey().equals("")){
			    System.out.println("error - key");
			}
			keyOut.writeBytes(e1[jx].getKey() + "\n");
			int[] fv = e1[jx].getFV();
			if (jx == 0){
			    System.out.println("num feats " + fv.length);
			}

			for(int kx = 0; kx < fv.length-1; kx++){
			    dataOut.writeBytes(fv[kx] +",");
			}
			fv = e2[jx].getFV();
			for(int kx = 0; kx < fv.length-1; kx++){
			    dataOut.writeBytes(fv[kx] +",");
			}

			numRules = fv.length-1;
			dataOut.writeBytes(fv[fv.length-1] + "\n");
		    }
		    else{
			System.err.println("error " + ix + " " + jx);
			System.err.println(e1[jx].getKey() + " " + e2[jx].getKey() + " " + e1[jx+1].getKey());
			
			System.exit(0);
		    }
		}
		for(int jx = 0; jx < (namesA.length-1); jx++){
		    newNames.writeBytes(namesA[jx] + "\n");
		}
		for(int jx = 0; jx < numRules; jx++){
		    newNames.writeBytes("r" + jx + " discrete 0,1\n");
		}
		newNames.writeBytes(namesA[namesA.length-1]);
		
		
		dataOut.close();
		keyOut.close();
		newNames.close();
		
	    }
	}
	catch(IOException exc){
	    System.out.println(exc.toString());
	    exc.printStackTrace();
	}
    }

    /**
       Converts the initial list of feature vectors into an array

       @param list a list of DataElement feature vectors
       @return an array of DataElement feature vectors
    */
    public Element[] convertListElements(LinkedList<Element> list){
	int len = list.size();
	Element[] elems = new Element[len];
	for(int ix = 0; ix < len; ix++){
	    elems[ix] = (Element)list.removeFirst();
	}
	Arrays.sort(elems);
	return elems;
    }

    /**
       Converts the initial list of feature vectors into an array

       @param list a list of DataElement feature vectors
       @return an array of DataElement feature vectors
    */
    public int[][] convertListInts(LinkedList<Element> list, Element[] elems){
	int len = list.size();
	int[][] examples = new int[len][];
	//elems = new Element[len];
	for(int ix = 0; ix < len; ix++){
	    elems[ix] = (Element)list.removeFirst();
	}
	Arrays.sort(elems);
	for(int ix = 0; ix < len; ix++){
	    examples[ix] = elems[ix].getFV();
	}
	return examples;
    }//end method convertListInts


    /**
       Converts the initial list of feature vectors into an array

       @param list a list of DataElement feature vectors
       @return an array of DataElement feature vectors
    */
    public int[][] convertListInts(LinkedList<Element> list){


	int len = list.size();
	int[][] examples = new int[len][];
	Element[] elems = new Element[len];
	//elems = new Element[len];
	for(int ix = 0; ix < len; ix++){
	    elems[ix] = (Element)list.removeFirst();
	}
	Arrays.sort(elems);
	for(int ix = 0; ix < len; ix++){
	    examples[ix] = elems[ix].getFV();
	}
	return examples;
    }//end method convertListInts

    public void readDataKey(String fileName, String keyFile, LinkedList<Element> exs, String sep){
	readDataKey(fileName, keyFile, exs, sep, false);
    }
	
    public void readDataKey(String fileName, String keyFile, LinkedList<Element> exs, String sep, boolean dynamic){
	try{
	    //System.out.println(fileName + " " + keyFile);
	    RandomAccessFile infile = new RandomAccessFile(fileName, "r");
	    RandomAccessFile infile2 = new RandomAccessFile(keyFile, "r");
	    String[] splt;
	    String[] keySplt;
	    String keyLine = infile2.readLine();
	    String line = infile.readLine();
	    int count = 1;
	    int cnt = 0;
	    while(line != null){
		cnt++;
		splt = line.split(sep);
		keySplt = keyLine.split(" ");
		int tmp[] = new int[splt.length];

		if  (splt.length > tmp.length){
		    System.out.println("error " + cnt + " " + splt.length);
		}
		for(int ix = 0; ix < splt.length; ix++){
		    try{
			tmp[ix] = Integer.parseInt(splt[ix]);
		    }
		    catch(NumberFormatException nfe){
			System.out.println("nfe " + cnt + " " + ix + " " + splt[ix].length() + " " + splt[ix]);
			if (splt[ix].equals("O")){
			    System.out.println("in here");
			    tmp[ix] = 0;
			}
			
		    }
		}//end of for loop
		try{
		    if (Integer.parseInt(keySplt[0]) == 271){
			//System.out.println("key 272 " + tmp[outputPosition]);
		    }
		}
		catch(NumberFormatException nfe){
		    System.out.println(count + " " + keyLine);
		}
		if (debug){
		    for(int ix = 0; ix < tmp.length; ix++){
			if (tmp[ix] < 0){
			    System.out.println("Error" + keyLine);
			}
		    }
		}
		exs.add(new Element(Integer.parseInt(keySplt[0]), keyLine, tmp, tmp[tmp.length-1]));
		keyLine = infile2.readLine();
		line = infile.readLine();

		count++;
	    }
	}//end try block
	catch(IOException exc){
	    System.out.println(exc.toString());
	    exc.printStackTrace();
	}
    }//read

    public Data(String names){
	nameRead(names);
    }

    /*
       reads in the <filestem>.names file and contructs a list of classes
       and a list of datapairs for the attributes

    */
    private void nameRead(String fileName) {
        int pos = 0;
        int vals = 0;
        Datapairs feature;
        Integer pairsInt;
        String check = "";

	LinkedList<Datapairs> names = new LinkedList<Datapairs>();
	LinkedList<String> attribValues = new LinkedList<String>();

	try{

	    StringTokenizer tokenizer;
	    RandomAccessFile input = new RandomAccessFile(fileName, "r");
	    while (input.getFilePointer() < input.length()) {  
		tokenizer = new StringTokenizer(input.readLine().trim(), " \n\r\t,");
		int count = tokenizer.countTokens();
		boolean done = false;
		String attribName;
		String attribType;
		if (count  >= 2){
		    attribName = tokenizer.nextToken();
		    attribType = tokenizer.nextToken();
		    //System.out.println(attribType);
		    if (attribType.equals("output"))
			outputPosition = names.size();
		    String test = "";
		    if (attribName.length() > 2){
			test = attribName.substring(0,2);
		    }
		    //System.out.println(attribName);
		    //System.out.println(test +" the string is " + attribName);
		    if (!test.equals("//")){
			//System.out.println(attribName);
			
			while ((tokenizer.hasMoreTokens()) && (!done)) {
			    
			    String token = tokenizer.nextToken();
			    token = token.trim();
			    if (token.length()  > 2)
				test = token.substring(0,2);
			    else
				test = token;
			    //System.out.println(test + " the string is " + attribName);
			    if (test.equals("//"))
				done = true;
			    else{
				//System.out.print(token + " ");
				attribValues.addLast(token);
			    }			    
			    //System.out.print(token + " ");
			}//while
			
			feature = new Datapairs(attribName, 
						attribValues.size(), attribType);

			//System.out.println(attribValues.size() + " num attribs");
			int listCount = attribValues.size();
			for (int i = 0; i < listCount; i++){
			    String temp = (String)attribValues.removeFirst();
			    //System.out.print(temp + " ");
			    feature.insertAttribVal(temp);
			}
			//System.out.println();
			//feature.insertAttribVal((String)attribValues.removeFirst());
			if (feature == null)
			    System.out.println("this is bad");
			names.add(feature);
		    }//end if(!attribName.equals("//"))
		    
		}//if count
	    }
	    input.close();
	}
	catch(IOException exc){
	    System.out.println(exc.toString());
	    exc.printStackTrace();
	    //System.out.println(exc.toString());
	}
      /*
          i think it is more efficient to store the Datapairs in an array
          also the order of the attributes was reversed as they were pulled
          out of the file, so they need to be switched
	*/
        dp = new Datapairs[names.size()];
	int dpCount = names.size();
        for (int ix = 0; ix < dpCount; ix++){
            dp[ix] = (Datapairs) names.removeFirst();
        }

    }//end of method
    public Data(){

    }

    public static void main(String[] args){
	Data data = new Data();
	//simpleJoin(String data1, String key1, String data2, String key2, String dataFile, String keyFile, String nameO, String nameN)
	//simpleJoinAgg(String path, String data2)
	data.simpleJoinAgg(args[0], args[1]);
	//data.simpleJoin(args[0], args[1], args[2],args[3], args[4], args[5], args[6], args[7] );//String data1, String key1, String data2, String key2, String dataFile, String keyFile, String nameO, String nameN)
    }
}
