/**
   @author Jesse Davis
   3/21/2004

   This class provides methods for implementing Nir Friedman's Maximum
   Likelihood Tree Augmented Naive Bayes Net.

 */
package sayu;
import java.util.*;

public class MaxTan{

    public static Boolean DEBUG = false;
    private static double initStrength = 5;
    //    public static int[] ignore = new int[4];
    /**
       calculates the conditional probability table for attribute at
       position pos in the feature.  The parent is supposed to the
       class variable.

       @param train - trainset
       @param pos - position of attribute in feature vector
       @param classPos - position of class variable in feature vector
       @param numVals - number of values of attribute
       @param numClass - number of classes
     */
    public static double[][] condCountOneAttrib(int[][] train, int pos, 
						int classPos, int numVals, 
						int numClass){
        //	System.out.println("Maxtan cond count");
        double[][] counts = new double[numVals][numClass];
        //for(int ix = 0; ix < counts.length; 
	double size = numVals * numClass;
        for(int ix = 0; ix < counts.length;ix++){
            for(int jx = 0; jx < counts[ix].length; jx++){
                counts[ix][jx] = initStrength/size;
            }
        }
        //calculate the counts
        for(int ix = 0; ix < train.length; ix++){
            if (train[ix][pos] >= counts.length){
                System.out.println("error" + pos + " " + numVals + " " + counts.length);
            }
            if (train[ix][pos] >= counts.length){
                System.out.println("this value error");
            }
            if (train[ix][pos] < 0){
                System.out.println("Very bad " + ix + " " + pos);
            }
            if (train[ix][classPos] >= counts[train[ix][pos]].length){
                System.out.println("class errror " + ix + " " + classPos + " " + train[ix][classPos]);
            }
            counts[train[ix][pos]][train[ix][classPos]]++;
        }
        double[][] probs = new double[counts.length][counts[0].length];
        for(int ix = 0; ix < counts[0].length;ix++){
            double total = 0;
            //get totals
            for(int jx = 0; jx < counts.length; jx++){
                total += counts[jx][ix];
            }
            //calculate the conditional probabilities
            for(int jx = 0; jx < counts.length; jx++){
                //System.out.print(counts[jx][ix] + " ");c
                probs[jx][ix] = (double)counts[jx][ix] / (double)total;
            }
            //  System.out.println(" / " + total);
        }//end of ix loop for getting totals/probs

        return probs;
    }//end of method condCountOneAttrib

    /**
       Calculates the counts for all combinations of attribute values
       between attributes at locations pos1, pos2 and class.

       @param pos1 - location of attribute 1 in feature the vector
       @param pos2 - location of attribute 2 in feature the vector
       @param train - the training set
       @param outputPos - location of class variable in feature the vector
       @param numVals1 - number of values for attribute 1
       @param numVals2 - number of values for attribute 2
       @param numClass - number of values for class variable

       @return joint distribution for these three attributes

     */
    public static double[][][] threeAttribCnts(int[][] train, int pos1, 
					       int pos2,
					       int classPos, int numVals1,
					       int numVals2, int numClass){

        double[][][] cnts = new double[numVals1][numVals2][numClass];
	double size = numVals1 * numVals2 * numClass;
        //initialize all counts to 1
        for(int ix = 0; ix < cnts.length;ix++){
            for(int jx = 0; jx <cnts[ix].length; jx++){
                for(int kx = 0; kx < cnts[ix][jx].length; kx++){
                    cnts[ix][jx][kx] = initStrength/size;
		    
                }
            }
        }
        //get the counts
        for(int ix = 0; ix < train.length; ix++){
            cnts[train[ix][pos1]][train[ix][pos2]][train[ix][classPos]]++;
        }
        return cnts;
    }//end of method threeAttribCnts


    /**
       Assumes that the last attribute is the class variable
       @param train - the training set
       @param outputPos - the position of the class variable in the
       feature vector
       @param attributes - description of all the attributes 

       @return a list of edges, where the weight on each edge is the
       conditional mutual information between the attributes the edges
       connects.
     */
    public static Edge[] getAllCondMutualInfo(int[][] train, int outputPos, 
					      VariableConsecInt[] attributes){

        /*
	  get the conditional distributions for all of the attributes 
	  in the network
         */
        double[][][] singleCondProbs = new double[attributes.length-1][][];

        //this allows for class variable not to be the last attribute
        int index = 0;
        int numClass = attributes[outputPos].getNumValues();
        int size = attributes.length-1;
        for(int ix = 0; ix < singleCondProbs.length; ix++){
            if (ix > outputPos){
                index = ix -1;
            }
            else{
                index = ix;
            }
            singleCondProbs[index] = condCountOneAttrib(train, ix, outputPos,
                    attributes[ix].getNumValues(), 
                    numClass);
        }//end of getting condition distributions

        //make a list of all possible edges
        Edge[] tree = new Edge[(size * (size-1))/2];
        int position = 0;
        //get the condition mutual information for all combinations of pairs
        for(int ix = 0; ix < (size-1); ix++){
            if (ix == outputPos){
                //ix++;
            }
            for(int jx = (ix+1); jx < size; jx++){  
                double ip = getIP(singleCondProbs[ix], singleCondProbs[jx],
                        ix, jx, train, outputPos,
                        attributes[ix].getNumValues(), 
                        attributes[jx].getNumValues(), numClass);
                tree[position] = new Edge(ix, jx, false, ip );
                position++;
            }//end of jx loop
        }//end of ix loop
        return tree;
    }//end of method getAllCondMutualInfo


    /**
       Calculates the conditional mutual information between two
       different attributes.

       @param attrib1 - the conditional probability table for attribute pos1
       with the class variable as its parent.
       @param attrib2 - the conditional probability table for attribute pos2
       with the class variable as its parent.
       @param pos1 - location of attribute 1 in feature the vector
       @param pos2 - location of attribute 2 in feature the vector
       @param train - the training set
       @param outputPos - location of class variable in feature the vector
       @param numVals1 - number of values for attribute 1
       @param numVals2 - number of values for attribute 2
       @param numClass - number of values for class variable
       @return the condition mutual information between the attribute
       at position pos1 and the attribute at position pos2
     */
    public static double getIP(double[][] attrib1, double[][] attrib2, 
            int pos1, int pos2, int[][] train, 
            int outputPos, int numVals1, int numVals2, 
            int numClass){


        //get joint distribution
        double[][][] cnts = threeAttribCnts(train, pos1, pos2, outputPos,
                numVals1, numVals2, numClass);

        double[] totals = new double[numClass];
        for(int ix = 0; ix < numClass; ix++){
            totals[ix] = 0;
        }
        double sumTot = 0;
        //get total and conditional totals over joint counts
        for(int ix = 0; ix < cnts.length; ix++){
            for(int jx = 0; jx < cnts[ix].length; jx++){
                for(int kx = 0; kx < cnts[ix][jx].length; kx++){
                    sumTot += cnts[ix][jx][kx];
                    totals[kx] += cnts[ix][jx][kx];
                }
            }
        }
	
        double sum = 0;
        double tmp1 = 0;
        double tmp2 = 0;
        double tmpSum = 0;
        //calculate the mutual information
        //System.out.println("MaxTan IP calc between " + pos1 + " " + pos2);
	//System.out.print("Scratch Totals: " + sumTot + " " + totals[0] + " " + totals[1]);
        for(int ix = 0; ix < cnts.length; ix++){
            for(int jx = 0; jx < cnts[ix].length; jx++){
                for(int kx = 0; kx < cnts[ix][jx].length; kx++){
                    //P(ix,jx,kx)
                    tmp1 = (double) cnts[ix][jx][kx]/(double)sumTot;
                    //P(ix,jx |kx)
                    tmp2 = (double) cnts[ix][jx][kx]/(double)totals[kx];
                    //attrib1[ix][kx] = P(ix | kx)
                    //attrib1[jx][kx] = P(jx | kx)
                    //System.out.print(tmp1 + " * " + tmp2 + "/(" + attrib1[ix][kx] + "*" + attrib2[jx][kx] + ") + ");
                    tmpSum = tmp1 * 
                    Math.log(tmp2 / (attrib1[ix][kx] * attrib2[jx][kx]));
                    sum += tmpSum;
                }//end of kx loop (all class values
            }//end of jx loop (all values for attribute at pos2
        }//end of ix loop (all values for attribute at pos1
	// System.out.println(" = " + sum);
        return sum;
    }//end of method getIP

    /**
       @param trainInts - the training set to draw the prob. distribution from
       @param outputPos - the positision in the feature vector that contains
                          the class variable.  the output positision must be
			  the last position in the feature vector.
       @return the maximum likelihood TAN structure for the trainInts
     */
    public static int[][] maxTanGraph(int[][] trainInts, int outputPos,
				      VariableConsecInt[] attributes, 
				      Edge[] pickedEdges){

        //get a list of all the edges in the graph
        Edge[] tree = getAllCondMutualInfo(trainInts,outputPos, attributes);
        int[] outDegrees = new int[attributes.length];
        //get a maximum weight spanning tree over the graph
        int[][] graph = kruskals(tree, outDegrees.length, outDegrees, pickedEdges);

        //make the graph directed. 
        makeDirected(graph, 0, outDegrees, -1);
 
        //connect the class node to all other nodes
        for(int ix = 0; ix < graph.length; ix++){
            if (ix != outputPos){
                graph[ix][outputPos] = Constants.IN_EDGE;
                graph[outputPos][ix] = Constants.OUT_EDGE;
            }
        }
        return graph;

    }//end of method maxTanGraph


    public static int[][] kruskals(Edge[] tree, int size, int[] outDegrees, Edge[] pickedEdges){
	return kruskals(tree,size,outDegrees,pickedEdges,0);
    }
    /**


     */
    public static int[][] kruskals(Edge[] tree, int size, int[] outDegrees, Edge[] pickedEdges, int ignoreSetSize){
        Arrays.sort(tree);
  
        int[][] graph = new int[size][size];
        //init graph to be empty
        for(int ix = 0; ix < size; ix++){
            for(int jx = 0; jx < size; jx++){
                graph[ix][jx] = Constants.NO_EDGE;
            }
            outDegrees[ix] = 0;
        }
        int v1, v2;
        //go through and try adding edges
        //	System.out.println("MaxTan kruskal - IP scores");
        double score =0;
        int curr = 0;
        for(int ix = (tree.length -1); ix >=0; ix--){
            //get information to add edge
            v1 = tree[ix].getStart();
            v2 = tree[ix].getEnd();
   
            //if (!Utils.contains(ignore, v1) && !Utils.contains(ignore, v2)){
            //add the edge
            graph[v1][v2] = Constants.IN_EDGE;
            graph[v2][v1] = Constants.IN_EDGE;
            outDegrees[v1]++;
            outDegrees[v2]++;
            //if add is invalid remove the edge
            if (!acyclic(graph, outDegrees, null, -1, false)){
		
                graph[v1][v2] = Constants.NO_EDGE;
                graph[v2][v1] = Constants.NO_EDGE;
                outDegrees[v1]--;
                outDegrees[v2]--;
            }
            else{
		/*
		System.out.print("TAN Scratch Picked Edge ");
		if (v1 < 10){
		    System.out.print(" ");
		}
		System.out.print(v1 + " ");
		if (v2 < 10){
		    System.out.print(" ");
		}
		System.out.println(v2 + " weight " + tree[ix].getWeight());
		*/	
                pickedEdges[curr] = new Edge(v1, v2, false, tree[ix].getWeight());
	       
                curr++;
                score+=tree[ix].getWeight();
                if (DEBUG){
                    //if (outDegrees.length == 20)
                    System.out.println("MaxTAN Pick: " + v1 + " " + v2 + " " + tree[ix].getWeight());
                }
            }
            //}
            //add edge to graph, see if still acylic
        }//end of ix loop
        //System.out.println("MaxTan weight " + score);
        //System.out.println("BN Weight: " + score);//weight);
        return graph;
    }//end of method kruskal

    /**
       turns the spanning tree into a directed graph. 
       @param graph is a undirected graph that representes a spanning tree
       @param root
       @param outDegrees
       @param parent

     */
    public static void makeDirected(int[][] graph, int root, int[] outDegrees,
            int parent){
        if (outDegrees[root] != 0){
            for(int ix = 0; ix < graph.length; ix++){
                if (graph[root][ix] != Constants.NO_EDGE &&
                        ix != parent){
                    outDegrees[ix]--;
                    graph[root][ix] = Constants.OUT_EDGE;
                    makeDirected(graph, ix, outDegrees, root);
                }
            }
        }
    }//end method makeDirected

    /**
       @param graph
       @param outDegrees
       @param 
       @return true if the graph is acyclic
     */
    public static boolean acyclic(int[][] graph, int[] outDegrees, 
            int[] inDegrees, int startNode, 
            boolean directed){
        if (startNode >= 0){
            if (inDegrees[startNode] == 0){
                return true;
            }
        }//
        int[][] tempGraph = (int[][])graph.clone();
        int vertex = tempGraph.length;
        boolean failed = false;
        boolean[] valid = new boolean[tempGraph.length];
        int[] tempOutDegrees = (int[]) outDegrees.clone();

        for (int ix = 0; ix < valid.length; ix++){
            valid[ix] = true;
        }
        while ((vertex > 0) && (!failed)){
            int pos = 0;
            boolean hasLeaf = false;
            for(; pos < tempOutDegrees.length && !hasLeaf; pos++){

                if (directed){
                    if ((tempOutDegrees[pos] == 0) && (valid[pos])){
                        hasLeaf = true;
                        vertex--;
                        valid[pos] = false;
                    }
                }
                else{
                    if ((tempOutDegrees[pos] <= 1 ) && (valid[pos])){
                        hasLeaf = true;
                        vertex--;
                        valid[pos] = false;
                    }
                }
            }//end of for loop

            if (!hasLeaf){
                //System.out.println("in here..why?");
                failed = true;
                continue;
            }
            /*
	      remove the vertex at pos from the graph and all the edges
	      entering the vertex
             */

            pos--;
            for(int i = 0; i <tempGraph.length && pos < tempGraph.length; i++){

                if (tempGraph[pos][i] == Constants.IN_EDGE){
                    tempOutDegrees[i ] = tempOutDegrees[i ] - 1;
                }
            }//end for
        }//end while

        return !failed;
    }//end method acyclic

    public static double[] getProbs(int[][] scoreSet, CondProb[] cpts){
        double cat1tmp = 1;
        double cat0tmp = 1;
        double norm, cat1Prob;

        double[] postProb = new double[scoreSet.length];

        for(int ix = 0; ix < scoreSet.length; ix++){
	    //Add in class prior
            cat1tmp = 0;
            cat0tmp = 0;
            //loop through the attributes
            for(int jx = 0; jx < cpts.length; jx++){
	
		cat1tmp = cat1tmp + cpts[jx].getLogProb(scoreSet[ix], true, 1);
                cat0tmp = cat0tmp + cpts[jx].getLogProb(scoreSet[ix], true, 0);
		/*
		if (ix == 1 && jx == 0){
		    System.out.print("M: ");
		}
		if (ix == 1){
System.out.print(jx + "(" + Utils.round(cpts[jx].getLogProb(scoreSet[ix], true, 1), 5)  + " " + Utils.round(cpts[jx].getLogProb(scoreSet[ix], true, 0), 5) + ") ");
		}
		
		if (ix == 0){
		    if (jx == 0){
			System.out.print("Scratch Ex: ");
			for(int xx = 0; xx < cpts.length; xx++){
			    System.out.print("(" + xx + ") " + scoreSet[ix][xx] + " ");
			}
			System.out.println();
			System.out.print("Scratch CPTs: ");
		    }
		    System.out.print(jx + "(" + Utils.round(cpts[jx].getLogProb(scoreSet[ix], true, 1), 5)  + " " + Utils.round(cpts[jx].getLogProb(scoreSet[ix], true, 0), 5) + ") ");
		}
		*/
            }//end of jx loop through individual feature vector
	    //get normalization constant
            norm = Math.exp(cat1tmp)+Math.exp(cat0tmp);
	    
	    //Calculate Probability
            cat1Prob = (Math.exp(cat1tmp) / norm);
            postProb[ix] = cat1Prob;
	    /*
	    if (ix == 1)
		System.out.println("prob = " + postProb[ix]);
	    */
        }
        return postProb;
    }



    public static CondProb[] buildCPTs(int[][] graph, Attribute[] attributes, int[][] data, int outputPos){
	CondProb[] cpts = new CondProb[graph.length];
	for(int ix = 0; ix < graph.length; ix++){
	    int[] parents = findParents(graph, ix);
	    cpts[ix] = new CondProb(ix, parents, outputPos, data, attributes);
	}
	return cpts;
    }

    public static int[] findParents(int[][] graph, int pos){
	int cnt = 0;
	for(int ix = 0; ix < graph.length; ix++){
	    if (graph[pos][ix] == Constants.IN_EDGE){
		cnt++;
	    }
	}
	int[] parents = new int[cnt];
	int index = 0;
	for(int ix = 0; ix < graph.length; ix++){
	    if (graph[pos][ix] == Constants.IN_EDGE){
		parents[index] = ix;
		index++;
	    }
	}
	return parents;
    }
}//end of class
