package sayu;
 
public class Utils{
    private static final double LN10 = Math.log(10);
    //code for rounding - from Adam Smith
    public static double round(double in, int nsig) {
        double pow;
        if (in == 0.0){
	    return 0.0;
	}
        pow = Math.pow(10.0, -(int)(Math.log(in) / LN10) +nsig);
        
	return Math.rint(in*pow)/pow;
    }
    

    public static void copyByteArray(byte[][] source, byte[][] target,
				     int skipPos){

	for(int ix = 0; ix < source.length; ix++){
	    int pos = 0;
	    for(int jx = 0; jx < source[ix].length; jx++){
		if (jx != skipPos){
		    target[ix][pos] = source[ix][jx];
		    pos++;
		}
	    }
	}
    }

   /**
       calculates the conditional probability table for attribute at
       position pos in the feature.  The parent is supposed to the
       class variable.

       @param train - trainset
       @param pos - position of attribute in feature vector
       @param classPos - position of class variable in feature vector
       @param numVals - number of values of attribute
       @param numClass - number of classes
     */
    public static double[][] condProb(byte[][] train, int pos, 
				      int classPos, int numVals, 
				      int numClass,
				      double initStrength, 
				      boolean logProbs){
        //	System.out.println("Maxtan cond count");
        double[][] counts = new double[numVals][numClass];
        //for(int ix = 0; ix < counts.length; 
	double size = numVals * numClass;
        for(int ix = 0; ix < counts.length;ix++){
            for(int jx = 0; jx < counts[ix].length; jx++){
                counts[ix][jx] = initStrength/size;
            }
        }
        //calculate the counts
        for(int ix = 0; ix < train.length; ix++){
            counts[train[ix][pos]][train[ix][classPos]]++;
        }
        double[][] probs = new double[counts.length][counts[0].length];
        for(int ix = 0; ix < counts[0].length;ix++){
            double total = 0;
            //get totals
            for(int jx = 0; jx < counts.length; jx++){
                total += counts[jx][ix];
            }
            //calculate the conditional probabilities
            for(int jx = 0; jx < counts.length; jx++){
		if (logProbs){
		    probs[jx][ix] = Math.log((double)counts[jx][ix] / (double)total);
		}
		else{
		    probs[jx][ix] = (double)counts[jx][ix] / (double)total;
		}
	    }
            //  System.out.println(" / " + total);
        }//end of ix loop for getting totals/probs

        return probs;
    }//end of method condCountOneAttrib

    
}