						%
% sayu feature evaluator
%
% all communication between JavaSayu and the rest of the world
% should be through this file.
%

:- module(cv_attributes, [cvsayu_remove_feature/2,
			    cvsayu_replace_feature/6,
			    cvsayu_new_feature/5,
			    cvsayu_current_test_score/1,
			    cvsayu_current_train_score/1,
			    cvsayu_commit_update/1,
			    cvsayu_mark_committed/1,
			    cvsayu_renew_object/0,
			    cvsayu_get_cuts/1,
			    cvsayu_covers_example/4,
			    cvsayu_new_search/1,
			    cvsayu_scoring_data/4,
			    cvsayu_init_scoring/1,
			    cvsayu_init/0,
			    cvsayu_exit/0,
			    cvsayu_current_score/1,
			    cvsayu_vals_to_scorable/2,
			    cvsayu_count_cover/2,
			    cvsayu_disjoin/3,
			    cvsayu_vals_to_scorable/3,
			    cvsayu_get_probs/2]).

:- use_module(scorer, [cur_likelihood/1]).


:- use_module('../common/utils',[randomize_list/2]).

:- use_module('../common/set',[vsc_set/1]).

:- use_module('../common/debug_write',[debug_format/3]).
%
% fold loading 
% 
:- use_module(load_folds,[get_folds/4]).

%
% reset computational object
% 
:- use_module(retrain,[retrain_model/0]).

:- use_module(library(lists),[sumlist/2]).

:- use_module(jpl_interface).

:- dynamic cvsayu_java_arrays/4.

:- dynamic cvsayu_object/1.

:- dynamic cvsayu_class/1.

cvsayu_class('sayu.CVSayu').

% scoring method:
% 0: weighted accuracy
% 1: Precision-Recall
% 2: AUC

:- dynamic cvsayu_scoring_method/1.

cvsayu_scoring_method(1).

% scoring method:
% 0: weighted accuracy
% 1: Precision-Recall
% 2: AUC
:- dynamic cvsayu_classifier_type/1.

cvsayu_classifier_type(1).

:- dynamic cvsayu_psuedo_count/1.

% psuedo count:
cvsayu_psuedo_count(1).

:- dynamic cvsayu_n_of_evaluated_atts/1.

cvsayu_n_of_evaluated_atts(0).

:- dynamic cvsayu_shuffle_training_set/1.

cvsayu_shuffle_training_set(at_commit).


% now permanently removes something out of network
cvsayu_remove_feature(Id, ScoreR) :-
	cvsayu_object(Object),
   	cvsayu_scoring_method(Method),
	debug_format(2,'-DJ- REPLACE ATTRIBUTE ~w~n',[Id]),
	myjpl_call(Object, removeAttribute, [Id, Method], ScoreR).

%
% tell SAYU that you don't want 
% 
cvsayu_replace_feature(Id, TrPos, TrNeg, TuPos, TuNeg, TePos, TeNeg, ScoreR) :-
	cvsayu_object(Obj),
   	cvsayu_scoring_method(Method),
	debug_format(1,'-DJ- REPLACE FEATURE ~w~n',[Id]),
	myjpl_call(Obj, replaceAttribute, [TrPos, TrNeg, TuPos, TuNeg, TePos, TeNeg, Id, Method, 2], ScoreR).

cvsayu_current_test_score(Score) :-
	cvsayu_object(SayuObj), 
	cvsayu_scoring_method(ScoringMethod),
	myjpl_call(SayuObj, getScore, [2,ScoringMethod], Score).

cvsayu_current_train_score(Score) :-
	cvsayu_object(SayuObj), 
	cvsayu_scoring_method(ScoringMethod),
	myjpl_call(SayuObj, getScore, [0,ScoringMethod], Score).

cvsayu_mark_committed(Score) :-
	cvsayu_object(SayuObj),
	myjpl_call(SayuObj, commit, [], _),
	retractall(cur_likelihood(_)),
	assert(cur_likelihood(Score)).

cvsayu_commit_update(Score) :-
	retractall(cur_likelihood(_)),
	assert(cur_likelihood(Score)),
	cvsayu_object(SayuObj),
	myjpl_call(SayuObj, commit, [], _),
	(
	 (
	   cvsayu_shuffle_training_set(at_commit)
	 ;
	   cvsayu_shuffle_training_set(at(_))
	 ), false
	->
	  shuffle_training_examples
	;
	  true
	),
%	cvsayu_scoring_method(ScoringMethod),
%	myjpl_call(Class, printCurve, [1,ScoringMethod], _),
%	myjpl_call(Class, printCurve, [2,ScoringMethod], _),
	cvsayu_dump_plist.

cvsayu_get_cuts(Thresholds) :-
	cvsayu_object(SayuObj),
	myjpl_call(SayuObj, getCuts, [], ThresholdsObj),
	jpl_array_to_list(ThresholdsObj, Thresholds).

cvsayu_new_feature(_,_,_,_, _) :-
	retract(cvsayu_n_of_evaluated_atts(Atts)),
	Atts1 is Atts+1,
	assert(cvsayu_n_of_evaluated_atts(Atts1)),
        cvsayu_shuffle_training_set(at(N)),
	0 =:= Atts mod N,
	shuffle_training_examples,
	fail.
cvsayu_new_feature(TrPos,TrNeg,TePos,TeNeg, Scores) :-
	cvsayu_scoring_method(Method),
	cvsayu_object(SayuObj),
	myjpl_call(SayuObj, addAttribute, [TrPos,TrNeg,TePos,TeNeg,Method], LikelihoodsObj),
	jpl_array_to_list(LikelihoodsObj, Scores).

cvsayu_covers_example(Obj, _PosNeg, _Type, Id) :-
	Indx is Id-1,
	jpl:jni_SetByteArrayElement(Obj, Indx, 1).

cvsayu_new_search(Num) :-
	cvsayu_object(SayuObj), 
	myjpl_call(SayuObj,newHead,[Num],_).
	
cvsayu_scoring_data(TrP,TrN,TeP,TeN) :-
	cvsayu_java_arrays(TrP,TrN,TeP,TeN).

cvsayu_init_scoring(Obj) :-
	myjpl_call( 'sayu.MyArrays', reset, [Obj], _).

cvsayu_current_score(Score) :-
	cur_likelihood(Score).

cvsayu_disjoin(Score, Rule1, Rule2) :-
	cvsayu_object(SayuObj),
	cvsayu_scoring_method(ScoringMethod),	
	%debug_format(1, 'calling disjoin~n', []),
	myjpl_call(SayuObj, disjoin, [ScoringMethod], Obj),
	%ThresholdsObj),
	jpl_array_to_list(Obj, [Score, Rule1, Rule2| _]).

cvsayu_count_cover(BMap, Cover) :-
	jpl_array_to_list(BMap,CoverList),
	xsum_list(CoverList,0,Cover).

xsum_list([],L,L).
xsum_list([0|L],I,IF) :-
	xsum_list(L,I,IF).
xsum_list([1|L],I,IF) :-
	I1 is I+1,
	xsum_list(L,I1,IF).


cvsayu_vals_to_scorable(Vals, Type, Obj) :-
	myjpl_list_to_array( Vals, Type, Obj).

cvsayu_get_probs(DataSet, AllProbs) :-
	cvsayu_object(Obj),
	myjpl_call(Obj, getProbs, [DataSet], Probs),
	jpl_array_to_list(Probs, AllProbs).

cvsayu_init :-
	cvsayu_do_init,
	retractall(cvsayu_n_of_evaluated_atts(_)),
	assert(cvsayu_n_of_evaluated_atts(0)).

cvsayu_do_init :-
%	myjpl_call('sayu.NBayes', init, [PSize,NSize], Likelihood),
	retractall(cur_likelihood(_)),
	retractall(cvsayu_java_arrays(_,_,_,_)),
	cvsayu_scoring_method(ScoringMethod),
%	cvsayu_psuedo_count(PsuedoCount),
	cvsayu_classifier_type(ClassifierType),
	get_folds(PosTrainFolds, NegTrainFolds, [TestPos], [TestNegs]),
	folds_to_examples(PosTrainFolds, 0, PosTrainExamples),
	folds_to_examples(NegTrainFolds, 0, NegTrainExamples),
	retractall(cvsayu_obj_settings(_, _, _, _, _)),
	length(PosTrainFolds, NFolds),
	myjpl_list_to_array(PosTrainExamples, int, PosDistArray),
	myjpl_list_to_array(NegTrainExamples, int, NegDistArray),
	assert(cvsayu_obj_settings(NFolds, PosDistArray, NegDistArray,TestPos,TestNegs)),
	cvsayu_class(SayuCV),
	myjpl_new(SayuCV,[ClassifierType,NFolds,PosDistArray,NegDistArray,TestPos,TestNegs], SayuObj),
	assert(cvsayu_object(SayuObj)),
	cvsayu_set_pr_area_threshold,
	myjpl_call(SayuObj, getScore, [2,ScoringMethod], InitialTestScore),
	myjpl_call(SayuObj, getScore, [0,ScoringMethod], InitialTrainScore),
	format(user_error,'Initial Train/Tune/Test Score: ~p/~p~n',[InitialTrainScore,InitialTestScore]),
	total_size(PosTrainFolds, 0, TotPos),
	total_size(NegTrainFolds, 0, TotNeg),
	zero_array(TotPos, PosArray),
	zero_array(TotNeg, NegArray),
	zero_array(TestPos, TestPosArray),
	zero_array(TestNegs, TestNegArray),
	assert(cvsayu_java_arrays(PosArray,NegArray,TestPosArray,TestNegArray)),
	myjpl_call(SayuObj, addAttribute, [PosArray,NegArray,TestPosArray,TestNegArray,ScoringMethod], LikelihoodsObj),
	jpl_array_to_list(LikelihoodsObj, Likelihoods),
	assert(cur_likelihood(Likelihoods)),
	assert(cvsayu_old_best(Likelihoods)).

total_size([], Tot, Tot).
total_size(L.PosTrainFolds, Tot0, Tot) :-
	TotI is Tot0+L,
	total_size(PosTrainFolds, TotI, Tot).

folds_to_examples([], _, []).
folds_to_examples(FoldSize.PosTrainFolds, F0, PosTrainExamples) :-
	F is F0+1,
	generate_examples(FoldSize, F0, PosTrainExamples, LeftPosTrainExamples),
	folds_to_examples(PosTrainFolds, F, LeftPosTrainExamples).

generate_examples(0, _, PosTrainExamples, PosTrainExamples) :- !.
generate_examples(FoldSize, F0, F0.PosTrainExamples, LeftPosTrainExamples) :-
	NFoldSize is FoldSize-1,
	generate_examples(NFoldSize, F0, PosTrainExamples, LeftPosTrainExamples).

cvsayu_set_pr_area_threshold :-
	user:sayu_min_recall_in_area(Threshold),
	myjpl_call('sayu.PrecisionRecallArea', setMinRecall, [Threshold], _).

cvsayu_renew_object :-
	cvsayu_obj_settings(NFolds, PosDistArray, NegDistArray,TestPos,TestNegs),
	cvsayu_class(SayuCV),
	cvsayu_classifier_type(ClassifierType),
	myjpl_new(SayuCV,[ClassifierType,NFolds,PosDistArray,NegDistArray,TestPos,TestNegs], SayuObj),
	retractall(cvsayu_object(_)),
	assert(cvsayu_object(SayuObj)).

cvsayu_exit :-
	flush_output,
	format(user_error,'% clauses in theory:~n',[]),
	fail.
cvsayu_exit :-
	recorded(theory_clause, Cl, _),
	portray_clause(user_error,Cl),
	fail.
cvsayu_exit :-
	cvsayu_object(Obj),
	cvsayu_scoring_method(ScoringMethod),
	myjpl_call(Obj, getScore, [2,ScoringMethod], TestLikelihood),
	format(user_error,'[ Final Test Score ~4f ]~n',[TestLikelihood]),
	fail.
cvsayu_exit :-
	% show current performance
	cvsayu_object(Obj),
	cvsayu_scoring_method(ScoringMethod),
	myjpl_call(Obj, printCurve, [2,ScoringMethod], _),
	fail.
cvsayu_exit :-
	% show current performance
%	cvsayu_class(Class),
%	cvsayu_scoring_method(ScoringMethod),
%	myjpl_call(Class, printCurve, [1,ScoringMethod], _),
%	myjpl_call(Class, printCurve, [2,ScoringMethod], _),
	fail.
cvsayu_exit :-
%	cvsayu_dump_plist,
	halt.

cvsayu_dump_plist :-
	( user:sayu_dump_file(F) -> open(F,write,S) ; S = user_error ),
	cvsayu_object(Obj),
	myjpl_call(Obj, getProbs, [2], Probs),
	once(current_predicate(_,test_pos:P)),
	findall(P,test_pos:P,LP),
	findall(P,test_neg:P,LN),
	jpl_array_to_list(Probs, AllProbs),
%	length(AllProbs,L0),
%	length(LP,LLP),
%	length(LN,LLN),
%	format(user_error,'probs: ~d ~d ~d ~n',[LLP,LLN,L0]),
	cvsayu_dump_probs(LN,neg,AllProbs,S,Rest),
	cvsayu_dump_probs(LP,pos,Rest,S,[]), !,
	close(S).

cvsayu_dump_probs([],_,Probs,_,Probs).
cvsayu_dump_probs([Ex|LP],Type,[P|AllProbs],S,Rest) :-
	format(S,'prob(~w,~q,~w).~n',[Type,Ex,P]),
	cvsayu_dump_probs(LP,Type,AllProbs,S,Rest).


shuffle_training_examples :-
	retract(sayu_obj_settings(NFolds, PosDistArray, NegDistArray,TestPos,TestNegs)),
	jpl_array_to_list(PosDistArray, PosDist),
	jpl_array_to_list(NegDistArray, NegDist),
	shuffle(PosDist, NFolds, NewPosDist),
	shuffle(NegDist, NFolds, NewNegDist),
	myjpl_list_to_array(NewPosDist, int, NewPosDistArray),
	myjpl_list_to_array(NewNegDist, int, NewNegDistArray),
	assert(cvsayu_obj_settings(NFolds, NewPosDistArray, NewNegDistArray,TestPos,TestNegs)),
	retrain_model.

shuffle([], _, []).
shuffle(_.PosDist, NFolds, N.NewPosDist) :-
	N is integer(random*NFolds),
	shuffle(PosDist, NFolds, NewPosDist).

	
