
:- module(load_folds, [get_folds/6,get_folds/4]).

% sayu_cv_folds or sayu_folds must be set before calling sayu.

:- dynamic sayu_cv_folds/4.

:- dynamic user:sayu_folds/3.


% first case: we need a split between train and tune. We assume if equal split.
get_folds(TrainPosSizes, TrainNegSizes, TunePosSizes, TuneNegSizes, TestPosSizes, TestNegSizes) :-
	sayu_cv_folds(TrainPosFiles,TrainNegFiles,TestPosFiles,TestNegFiles), !,
	get_folds(TrainPosFiles, train_pos, tune_pos, TrainPosSizes, TunePosSizes),
	get_folds(TrainNegFiles, train_neg, tune_pos, TrainNegSizes, TuneNegSizes),
	get_folds(TestPosFiles, test_pos, TestPosSizes),
	get_folds(TestNegFiles, test_neg, TestNegSizes).

get_folds(TrainPosSizes, TrainNegSizes, TunePosSizes, TuneNegSizes, TestPosSizes, TestNegSizes) :-
	% old interface, everything is in sayu_prefix
	user:sayu_prefix(Prefix),
	get_unqualified_folds(Prefix, train, TrainPosSizes, TrainNegSizes),
	get_unqualified_folds(Prefix, tune, TunePosSizes, TuneNegSizes),
	get_unqualified_folds(Prefix, test, TestPosSizes, TestNegSizes).

get_folds(Files, TR, TU, TRS, TUS) :-
	length(Files, NFolds),
	TRFolds is NFolds//2,
	split_list(TRFolds, Files, TRFiles, TUFiles),
	get_folds(TRFiles, TR, TRS),
	get_folds(TUFiles, TU, TUS).

split_list(0, Files, [], Files) :- !.
split_list(TRFolds, F.Files, F.TRFiles, TUFiles) :-
	TRFolds1 is TRFolds-1,
	split_list(TRFolds1, Files, TRFiles, TUFiles).

get_folds([], _, []).
get_folds(F.TrainPosFiles, TypeFold, Size.TrainPosSizes) :-
	load_files(temp:F,[silent(true)]),
	once(current_predicate(_,temp:P)),
	predicate_property(temp:P, number_of_clauses(Size)),
	add_examples(P, TypeFold), 
	get_folds(TrainPosFiles, TypeFold, TrainPosSizes).

get_folds(TrainPosSizes, TrainNegSizes, TestPosSizes, TestNegSizes) :-
	sayu_cv_folds(TrainPosFiles,TrainNegFiles,TestPosFiles,TestNegFiles),
	get_folds(TrainPosFiles, train_pos, TrainPosSizes),
	get_folds(TrainNegFiles, train_neg, TrainNegSizes),
	get_folds(TestPosFiles, test_pos, TestPosSizes),
	get_folds(TestNegFiles, test_neg, TestNegSizes).

get_unqualified_folds(IPrefix, TFold, [PSize], [NSize]) :-
	atom_concat([IPrefix,TFold,'.f'],PF),
	atom_concat(TFold,'_pos',PTypeFold),
	load_files(temp:PF,[silent(true)]),
	once(current_predicate(_,temp:P)),
	predicate_property(temp:P, number_of_clauses(PSize)),
	add_examples(P, PTypeFold), 
	atom_concat([IPrefix,TFold,'.n'],NF),
	atom_concat(TFold,'_neg',NTypeFold),
	load_files(temp:NF,[silent(true)]),
	predicate_property(temp:P, number_of_clauses(NSize)),
	add_examples(P, NTypeFold).

add_examples(P, TypeFold) :-
	call(temp:P),
	assertz(TypeFold:P),
	fail.
add_examples(_, _).

