%
% weka feature evaluator
%
% all communication between JavaWeka and the rest of the world
% should be through this file.
%

:- module(weka_attributes, [weka_remove_feature/2,
			    weka_replace_feature/6,
			    weka_new_feature/5,
			    weka_current_test_score/1,
			    weka_current_train_score/1,
			    weka_commit_update/1,
			    weka_mark_committed/1,
			    weka_get_cuts/1,
			    weka_covers_example/4,
			    weka_new_search/1,
			    weka_scoring_data/4,
			    weka_init_scoring/1,
			    weka_init/0,
			    weka_exit/0,
			    weka_current_score/1,
			    weka_vals_to_scorable/2,
			    weka_count_cover/2,
			    weka_disjoin/3,
			    weka_renew_object/0,
			    weka_vals_to_scorable/3,
			    weka_get_probs/2]).

:- use_module(scorer, [cur_likelihood/1]).

:- use_module('../common/utils',[randomize_list/2]).

:- use_module('../common/set',[vsc_set/1]).

:- use_module('../common/debug_write',[debug_format/3]).

%
% fold loading 
% 
:- use_module(load_folds,[get_folds/6]).

%
% reset computational object
% 
:- use_module(retrain,[retrain_model/0]).

:- use_module(library(lists),[sumlist/2]).

:- use_module(jpl_interface).

:- dynamic weka_java_arrays/6.

:- dynamic weka_java_sizes/3.

:- dynamic weka_object/1.

:- dynamic weka_class/1.

weka_class('sayu.WekaClassifier').

% scoring method:
% 0: weighted accuracy
% 1: Precision-Recall
% 2: AUC

:- dynamic weka_scoring_method/1.

:- vsc_set(weka_scoring_method(1)).

% Classifier type
% currently only jtree
:- dynamic weka_classifier_type/1.

weka_classifier_type(4).

% Classifier options
% may differ for each classifier
% no options right now
:- dynamic weka_classifier_options/2.

weka_classifier_options(1,'-C 0.1 -M 2').
weka_classifier_options(2,'-I 10 -K 0 -S 1').
weka_classifier_options(3,'-I 10 -K 0 -S 1').
weka_classifier_options(4,'').
weka_classifier_options(5,'-Q weka.classifiers.bayes.net.search.local.TAN -- -S BDeu -E weka.classifiers.bayes.net.estimate.SimpleEstimator -- -A 4.0').
weka_classifier_options(6,'-cost-matrix "[0.0 1.0; 100.0 0.0]" -M -S1 -W weka.classifiers.trees.Id3').
%weka_classifier_options(6,'-cost-matrix "[0.0 100.0; 1.0 0.0]" -M -S1 -W weka.classifiers.functions.SimpleLogistic').

:- dynamic weka_n_of_evaluated_atts/1.

weka_n_of_evaluated_atts(0).

%
% tell WEKA that you don't want 
% 
weka_remove_feature(Id, Score) :-
	weka_object(Object),
   	weka_scoring_method(Method),
	myjpl_call(Object, removeFeature, [Id,Method], Score).

weka_replace_feature(Id, TrPos-TuPos, TrNeg-TuNeg, TePos, TeNeg, ScoreR) :-
	weka_object(Object),
   	weka_scoring_method(Method),
	myjpl_call(Object, replaceAttribute, [TrPos, TrNeg, TuPos, TuNeg, TePos, TeNeg, Id, Method, 2], ScoreR).

weka_current_test_score(Score) :-
	weka_object(WekaObj), 
	weka_scoring_method(ScoringMethod),
	myjpl_call(WekaObj, getScore, [2,ScoringMethod], Score).

weka_current_train_score(Score) :-
	weka_object(WekaObj), 
	weka_scoring_method(ScoringMethod),
	myjpl_call(WekaObj, getScore, [0,ScoringMethod], Score).

weka_mark_committed(Score) :-
	weka_object(WekaObj),
	myjpl_call(WekaObj, commit, [], _),
	retractall(cur_likelihood(_)),
	assert(cur_likelihood(Score)).

weka_commit_update(Score) :-
	retractall(cur_likelihood(_)),
	assert(cur_likelihood(Score)),
	weka_object(WekaObj),
	myjpl_call(WekaObj, commit, [], _),
%	weka_scoring_method(ScoringMethod),
%	myjpl_call(Class, printCurve, [1,ScoringMethod], _),
%	myjpl_call(Class, printCurve, [2,ScoringMethod], _),
	weka_dump_plist.

weka_get_cuts(Thresholds) :-
	weka_object(WekaObj),
	myjpl_call(WekaObj, getCuts, [], ThresholdsObj),
	jpl_array_to_list(ThresholdsObj, Thresholds).

weka_new_feature(_,_,_,_, _) :-
	retract(weka_n_of_evaluated_atts(Atts)),
	Atts1 is Atts+1,
	assert(weka_n_of_evaluated_atts(Atts1)),
	fail.
weka_new_feature(TrPos-TuPos,TrNeg-TuNeg,TePos,TeNeg, Likelihood) :-
	weka_scoring_method(Method),
	weka_object(WekaObj),
	myjpl_call(WekaObj, newattribute, [TrPos,TrNeg,TuPos,TuNeg,TePos,TeNeg,Method], Likelihood).

weka_covers_example(TrainObj-TuneObj, PosNeg, _Type, Id) :- !,
	weka_java_sizes(train, PosNeg, TrainPosSize ),
	(
	  Id > TrainPosSize
	->
	  Indx is Id-(TrainPosSize+1),
	  jpl:jni_SetByteArrayElement(TuneObj, Indx, 1)
	;
	  Indx is Id-1,
	  jpl:jni_SetByteArrayElement(TrainObj, Indx, 1)
	).
weka_covers_example(Obj, _PosNeg, _Type, Id) :-
	Indx is Id-1,
	jpl:jni_SetByteArrayElement(Obj, Indx, 1).

weka_new_search(Num) :-
	weka_object(WekaObj), 
	myjpl_call(WekaObj,newHead,[Num],_).
	
weka_scoring_data(TrP-TuP,TrN-TuN,TeP,TeN) :-
	weka_java_arrays(TrP,TrN,TuP,TuN,TeP,TeN).

weka_init_scoring(Obj1-Obj2) :- !,
	myjpl_call( 'sayu.MyArrays', reset, [Obj1], _),
	myjpl_call( 'sayu.MyArrays', reset, [Obj2], _).
weka_init_scoring(Obj) :-
	myjpl_call( 'sayu.MyArrays', reset, [Obj], _).

weka_current_score(Score) :-
	cur_likelihood(Score).

weka_disjoin(Score, Rule1, Rule2) :-
	weka_object(WekaObj),
	weka_scoring_method(ScoringMethod),	
	%debug_format(1, 'calling disjoin~n', []),
	myjpl_call(WekaObj, disjoin, [ScoringMethod], Obj),
	%ThresholdsObj),
	jpl_array_to_list(Obj, [Score, Rule1, Rule2| _]).

weka_count_cover(BMap, Cover) :-
	jpl_array_to_list(BMap,CoverList),
	xsum_list(CoverList,0,Cover).

xsum_list([],L,L).
xsum_list([0|L],I,IF) :-
	xsum_list(L,I,IF).
xsum_list([1|L],I,IF) :-
	I1 is I+1,
	xsum_list(L,I1,IF).


weka_vals_to_scorable(Vals, Type, Obj) :-
	jpl_list_to_array( Vals, Type, Obj).

weka_get_probs(DataSet, AllProbs) :-
	weka_object(Obj),
	myjpl_call(Obj, getProbs, [DataSet], Probs),
	jpl_array_to_list(Probs, AllProbs).

weka_init :-
	weka_init_jvm,
	weka_do_init,
	retractall(weka_n_of_evaluated_atts(_)),
	assert(weka_n_of_evaluated_atts(0)).

weka_init_jvm :-
	predicate_property(jpl:myjpl_call(_,_,_,_),_), !.
weka_init_jvm :-
	weka_add_weka_classes.

:- prolog_load_context(directory,Dir),
	atom_concat([Dir,'/../sayu.jar:',Dir,'/../weka.jar'],JPath),
	assert_static(weka_location(JPath)).

% for experimenting with hierarchy learning
weka_add_weka_classes :-
	unix(environ('CLASSPATH',CP)), !,
	weka_location(JPath),
	atom_concat([JPath,':',CP],NEWCLASSPATH),
	unix(putenv('CLASSPATH',NEWCLASSPATH)).
weka_add_weka_classes :- 
	weka_location(JPath),
	unix(putenv('CLASSPATH',JPath)).

weka_do_init :-
%	myjpl_call('weka.NBayes', init, [PSize,NSize], Likelihood),
	retractall(cur_likelihood(_)),
	retractall(weka_java_arrays(_,_,_,_,_,_)),
	retractall(weka_java_sizes(_,_,_)),
	weka_scoring_method(ScoringMethod),
	weka_classifier_type(ClassifierType),
	weka_classifier_options(ClassifierType, ClassifierOpts),
	get_folds(PosTrainFolds, NegTrainFolds, PosTuneFolds, NegTuneFolds, [TestPos], [TestNegs]),
	sumlist(PosTrainFolds, TrainPos),
	sumlist(PosTuneFolds, TunePos),
	sumlist(NegTrainFolds, TrainNeg),
	sumlist(NegTuneFolds, TuneNeg),
	retractall(weka_obj_settings(_, _, _, _)),
	assert(weka_obj_settings(TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs)),
	weka_class(WekaClass),
	myjpl_new(WekaClass,[ClassifierType, TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs, ClassifierOpts], WekaObj),
	assert(weka_object(WekaObj)),
	weka_set_pr_area_threshold,
	writeln(done(set)),
	myjpl_call(WekaObj, getScore, [2,ScoringMethod], InitialTestScore),
	myjpl_call(WekaObj, getScore, [1,ScoringMethod], InitialTuneScore),
	myjpl_call(WekaObj, getScore, [0,ScoringMethod], InitialTrainScore),
	format(user_error,'Initial Train/Test Score: ~p/~p/~p~n',[InitialTrainScore,InitialTuneScore,InitialTestScore]),
	zero_array(TrainPos, TrainPosArray),
	zero_array(TrainNeg, TrainNegArray),
	zero_array(TunePos, TunePosArray),
	zero_array(TuneNeg, TuneNegArray),
	zero_array(TestPos, TestPosArray),
	zero_array(TestNegs, TestNegArray),
	assert(weka_java_arrays(TrainPosArray,TrainNegArray,TunePosArray,TuneNegArray,TestPosArray,TestNegArray)),
	assert(weka_java_sizes(train, pos, TrainPos )),
	assert(weka_java_sizes(train, neg, TrainNeg )),
	assert(weka_java_sizes(tune, pos ,TunePos)),
	assert(weka_java_sizes(tune, neg, TuneNeg)),
	assert(weka_java_sizes(test, pos, TestPos)),
	assert(weka_java_sizes(test, neg, TestNegs)),
	myjpl_call(WekaObj, newattribute, [TrainPosArray,TrainNegArray,TunePosArray,TuneNegArray,TestPosArray,TestNegArray,ScoringMethod], Likelihood),
	assert(cur_likelihood(Likelihood)),
	assert(weka_old_best(Likelihood)).

total_size([], Tot, Tot).
total_size(L.PosTrainFolds, Tot0, Tot) :-
	TotI is Tot0+L,
	total_size(PosTrainFolds, TotI, Tot).

weka_exit :-
	flush_output,
	format(user_error,'% clauses in theory:~n',[]),
	fail.
weka_exit :-
	recorded(theory_clause, Cl, _),
	portray_clause(user_error,Cl),
	fail.
weka_exit :-
	weka_object(Obj),
	weka_scoring_method(ScoringMethod),
	myjpl_call(Obj, getScore, [2,ScoringMethod], TestLikelihood),
	format(user_error,'[ Final Test Score ~4f ]~n',[TestLikelihood]),
	fail.
weka_exit :-
	% show current performance
	weka_object(Obj),
	weka_scoring_method(ScoringMethod),
	myjpl_call(Obj, printCurve, [2,ScoringMethod], _),
	fail.
weka_exit :-
	% show current performance
%	weka_class(Class),
%	weka_scoring_method(ScoringMethod),
%	myjpl_call(Class, printCurve, [1,ScoringMethod], _),
%	myjpl_call(Class, printCurve, [2,ScoringMethod], _),
	fail.
weka_exit :-
%	weka_dump_plist,
	halt.

weka_dump_plist :-
	( user:sayu_dump_file(F) -> open(F,write,S) ; S = user_error ),
	weka_object(Obj),
	myjpl_call(Obj, getProbs, [2], Probs),
	once(current_predicate(_,test_pos:P)),
	findall(P,test_pos:P,LP),
	findall(P,test_neg:P,LN),
	jpl_array_to_list(Probs, AllProbs),
%	length(AllProbs,L0),
%	length(LP,LLP),
%	length(LN,LLN),
%	format(user_error,'probs: ~d ~d ~d ~n',[LLP,LLN,L0]),
	weka_dump_probs(LN,neg,AllProbs,S,Rest),
	weka_dump_probs(LP,pos,Rest,S,[]), !,
	close(S).

weka_dump_probs([],_,Probs,_,Probs).
weka_dump_probs([Ex|LP],Type,[P|AllProbs],S,Rest) :-
	format(S,'prob(~w,~q,~w).~n',[Type,Ex,P]),
	weka_dump_probs(LP,Type,AllProbs,S,Rest).


weka_set_pr_area_threshold :-
	user:sayu_min_recall_in_area(Threshold),
	myjpl_call('sayu.PrecisionRecallArea', setMinRecall, [Threshold], _).

weka_renew_object :-
	sayu_obj_settings(TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs),
	weka_class(WekaClass),
	weka_classifier_type(ClassifierType),
	weka_classifier_options(ClassifierType, ClassifierOpts),
	myjpl_new(WekaClass,[ClassifierType, TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs, ClassifierOpts], WekaObj),
	retractall(weka_object(_)),
	assert(weka_object(WekaObj)).

