%
%

:- module(vista_correlation,[
	correlated_constant/7,
	compute_counts/1,
	extra_constant/4]).

:- use_module(bottom_clause,[
	constant_histogram/0,
	constant_in_histogram/3,
	hist/4,
	has_hist/0]).

:- use_module(hierarchy,[
	ok_type/1]).

:- use_module(library(lists),[nth/3,member/2,sumlist/2]).

:- dynamic 
	counts_known/0,
	co/3,
	count/4,
	ec/4.

extra_constant(Cs, Min, Type, Constant) :-
	ec(Cs, Min, Type, _), !,
	ec(Cs, Min, Type, Constant).
extra_constant(Cs, Min, Type, Constant) :-
	setof(Corr-Constant,correlated_constant(Cs, Type, Min, phi, 0.2, Constant, Corr), L),
%	length(L, N),
%	cut_above(20,N,L,Members),
	L = Members,
	member(_-Constant, Members),
	assert(ec(Cs, Min, Type, Constant)).

cut_above(N0,N,L,L) :- N =< N0, !.
cut_above(N0,N,_.L,Members) :-
	N1 is N-1,
	cut_above(N0,N1,L,Members).

correlated_constant(Cs, Type, Min, Correlation, MinCorr, Constant, Corr) :-
	member(C1, Cs),
	constant_in_histogram(Type, Min, Constant),
	C1 @< Constant,
	correlation(C1, Constant, Type, Min, Corr, Correlation),
	abs(Corr) >= MinCorr.

correlation(C1, Constant, Type, Min, Corr, pearson) :-
	setof(C, constant_in_histogram(Type, Min, C), Consts),
	constant_counts(Consts, Type), 
	length(Consts, N),
	pearson_correlation(C1, Constant, N, Type, Min, Corr).
correlation(C1, Constant, Type, Min, Corr, phi) :-
	phi_correlation(C1, Constant, Type, Min, Corr).

compute_counts(_Min) :-
	counts_known, !.
compute_counts(Min) :-
	has_hist, !,
	generate_counts(Min),
	assert(counts_known).
compute_counts(Min) :-
	constant_histogram,
	compute_counts(Min).
	
generate_counts(Min) :-
%	PosNeg = pos,
	example(Id, PosNeg, Ex),
	\+ ( example(Id1, PosNeg, Ex), Id1 < Id),	
	sat(Id, PosNeg),
	add_to_correlation_count(Min, PosNeg),
	fail.
generate_counts(_).

add_to_correlation_count(Min, Polarity) :-
	ok_type(Type),
	common(C1, Type, C2),
	C1 @=< C2,
	hist(Type, C1, Polarity, N1),
 	N1 >= Min,
	hist(Type, C2, Polarity, N2),
	N2 >= Min,
	inc_count(C1, Type, C2),
%writeln(x:C1:C2),
	fail.

common(C1, Type, C2) :-
	modeb(Mode),
	Mode =.. [Na|Args],
	two_goals(Args, Type, C1, C2, G1L, G2L),
	G1 =.. [Na|G1L],
	G2 =.. [Na|G2L],
	% this is where the backtracking happens...
	bc(_,_,G1),
	bc(_,_,G2),
	G2 \= G1.

two_goals(#T.Args, T, C1, C2, C1.G1L, C2.G2L) :- !,
	two_goalsb(Args, G1L, G2L).
two_goals(+_.Args, T, C1, C2, A.G1L, A.G2L) :- !,
	two_goals(Args, T, C1, C2, G1L, G2L).
two_goals(_.Args, T, C1, C2, _.G1L, _.G2L) :-
	two_goals(Args, T, C1, C2, G1L, G2L).

% old version
%two_goals((_).Args, T, C1, C2, A.G1L, A.G2L) :-
%	two_goals(Args, T, C1, C2, G1L, G2L).
% two_goals((-_).Args, T, C1, C2, _.G1L, _.G2L) :-
% 	two_goals(Args, T, C1, C2, G1L, G2L).

two_goalsb([], [], []).
two_goalsb(_.Args, _.G1L, _.G2L) :-
	two_goalsb(Args, G1L, G2L).

%inc_count(C1, _Type, C2) :-
%	found(C1,C2), !.
inc_count(C1, _Type, C2) :-
%	assert(found(C1,C2)),
	retract(co(C1,C2,N)), !,
	N1 is N+1,
	assert(co(C1,C2,N1)).
inc_count(C1, _Type, C2) :-
	assert(co(C1,C2,1)).

phi_correlation(C1, C2, Type, Min, Corr) :-
	C00 = c00(0),
	C01 = c01(0),
	C10 = c10(0),
	C11 = c11(0),
	phi_correlation(C1, C2, Type, Min, Corr, C00, C01, C10, C11).

phi_correlation(C1, C2, Type, Min, _, C00, C01, C10, C11) :-
	constant_in_histogram(Type, Min, C),
	co_2(C1,C,X1),
	co_2(C2,C,X2),
	(X1 =:= 0
	->
	  ( X2 =:= 0 -> inc(C00) ; inc(C01) )
	  ;
	  ( X2 =:= 0 -> inc(C10) ; inc(C11) )
        ),
	fail.
phi_correlation(_, _, _, _, Corr, coo(N00), c01(N01), c10(N10), c11(N11)) :-
	N11 \= 0,
	Corr is (N11*N00-N01*N10)/sqrt((N00+N01)*(N10+N11)*(N00+N10)*(N01+N11)).

inc(C) :-
	arg(1, C,V),
	V1 is V+1,
	nb_setarg(1,C,V1).

constant_counts(_, Type) :-
	count(_,Type,_,_), !.
constant_counts(Consts,Type) :-
	constant_counts_rec(Consts,Type).

constant_counts_rec([], _).
constant_counts_rec(C.Consts, Type) :-
	x_xx(C, Count, Count2),
	assert(count(C, Type, Count, Count2)),
	constant_counts_rec(Consts, Type).

x_xx(C, Count, Count2) :-
	findall(N, co2(C,_,N), Ls),
	sumlist(Ls, Count),
	sumsquares(Ls, 0, Count2).

co2(C,C1,N) :-
	co(C,C1,N).
co2(C,C1,N) :-
	co(C1,C,N).

sumsquares([], N, N).
sumsquares(El.L, N0, N) :-
	N1 is N0+El*El,
	sumsquares(L, N1, N).

pearson_correlation(C1, C2, N, Type, _Min, Corr) :-
	findall(Count, xy(C1,C2,_,Count), L),
	sumlist(L, CXY),
	count(C1, Type, CX, CXX),
	count(C2, Type, CY, CYY),
	Den is (sqrt(N*CXX-CX*CX)*sqrt(N*CYY-CY*CY)),
	Num is  (N*CXY-CX*CY),
	(Num =:= 0 -> Corr = 0.0 ; Corr is Num/Den).

xy(C1,C2,C,Count) :-
	co(C1,C,X),
	( co(C2,C,Y) -> true ; co(C,C2,Y) ),
	Count is X*Y.
xy(C1,C2,C,Count) :-
	co(C,C1,X),
	( co(C2,C,Y) -> true ; co(C,C2,Y) ),
	Count is X*Y.

co_1(C1,C,X) :-
	C1 @=< C, !,
	( co(C1,C,X) -> true ; X = 0).
co_1(C1,C,X) :-
	( co(C,C1,X) -> true ; X = 0).

co_2(C1,C,X) :-
	( C1 @=< C -> co(C1,C,X) ; co(C,C1,X) ).


add(C,V) :-
	arg(1,C,V0),
	V1 is V0+V,
	nb_setarg(1,C,V1).


