
:- module(vista_hierarchy, [reduce_hierarchy/0,
	build_hierarchy/1,
	ok_type/1,
	hier_clause/8,
	commit_hierarchy/2]).

:- use_module('../common/debug_write', [debug_format/3]).

:- use_module(variabilise,[unvariabilise/2]).

:- use_module(theory,[theory_varclause/2,
		  replace_clause_in_theory/3,
		  list_theory/0]).

:- use_module(clause,[has_constant/1,
		  replace_constant/5]).

:- use_module(eval_node,[store_final_clause/6]).

:- use_module(examples,[new_predicate_to_call/1]).

:- use_module(score,[score_replacement/7]).

:- use_module(aggregation,[aggregation/3]).

:- use_module(constants,[get_constant/4]).

:- use_module(newpred,[get_const_pred_name/1]).

:- use_module('../scorer/scorer', [commit_update/1,
				   better_score/2,
				   better_hierarchy_score/2,
				   increased_score/2,
				   current_score/1,
				   current_test_score/1]).

:- use_module(library(lists),[delete/3]).

:- dynamic hierarchy_ok/1, hierarchy_predicate/3, h/1, hierarchy_stratified/1, found/3.

hierarchy_ok(all_types).

hierarchy_stratified(true).

reduce_hierarchy :-
	reduce_hierarchy(1).

reduce_hierarchy(LowClause) :-
	current_score(Score0),
	theory_varclause(Id, (Head :- Body)),
	Id >= LowClause,
	hier_clause((Head :- Body), Score0, Id, +inf, single, Score, NewClause, _),
writeln(new:NewClause),
        better_hierarchy_score(Score, Score0),
	!,
	debug_format(1,'-DJ- HIER FOUND~n',[]),
	complete_search(NewClause, Id, Score).
reduce_hierarchy(_) :-
	debug_format(1,'-DJ- NO HIER FOUND~n',[]).

hier_clause((Head :- Body), Score0, Id, TPos, _UseAggregates, Score, NewClause, Type) :- 
	has_constant(Body), !,
	mk_disj(Head, Body, Id, TPos, Score0, Score, NewClause, Type).
hier_clause((Head :- Body), Score0, Id, TPos, _UseAggregates, Score, (Head:-Body), Type) :- 
	has_hierarchy(Body, Type, Name),
	add_to_disj(Head, Body, Id, Type, Name, TPos, Score0, Score).

mk_disj(Head, Body, Id, TPos, Score, NewScore, NewClause, Type) :-
	replace_constant(Body, Const, NewBody, Var, #Type),
	ok_type(Type),
	construct_disj(Head, NewBody, Const, Var, Id, TPos, Type, NewClause, Score, NewScore).

construct_disj(Head, NewBody, Const, NewVar, Id, TPos, CType, NewClause, OldScore, NewScore) :-
	NewClause = (Head :- G,NewBody),
	get_const_pred_name(Name),
	G =.. [Name, NewVar],
	new_hierarchy_predicate_to_call(G, CType),
	G0 =.. [Name, Const],
%	debug_format(2,'-DJ- FAC 4a~w/~w.~n',[G0,OldScore]),
	assert(user:G0),
	Type = t,
	UseAggregates = single,
	aggregation(NewClause, AggregatedClause, UseAggregates),
	aggregation((Head :- NewBody), SoftClause, UseAggregates),
	expand_disj(AggregatedClause, SoftClause/NewVar, [Const], CType, Name, Id, TPos, Type, UseAggregates, OldScore, NewScore).


add_to_disj(Head, (Disj,Body), Id, CType, Name, TPos, OldScore, NewScore) :-
	get_consts(Disj, Name, Consts),
	debug_format(1,'-DJ- FAC EXPAND HIER~w/~w.~n',[Disj,OldScore]),
	Type = t,
	UseAggregates = single,
	aggregation((Head :- (Disj,Body)), AggregatedClause, UseAggregates),
	aggregation((Head :- Body), SoftClause, UseAggregates),
	arg(1, Disj,Var),
	expand_disj(AggregatedClause, SoftClause/Var, Consts, CType, Name, Id, TPos, Type, UseAggregates, OldScore, NewScore).

expand_disj(NewClause, TopClause, Consts, CType, CName, Id, TPos, Type, UseAgg, OldScore, NewScore) :-
	NewGoal =.. [CName, X],
	findall(OutGoal,call_constant(Consts, TopClause, CType, NewGoal, X, OutGoal),ConstGoals),
	ConstGoals = [_|_],
	debug_format(1,'-DJ- EXPAND HIER~w/~w.~n',[Consts,OldScore]),
	add_consts(ConstGoals, Consts, TopClause, CType, NewClause, Id, Type, TPos, UseAgg, OldScore, NewScore).

call_constant(Consts, TopClause, CType, NewGoal, Constant, OutGoal) :-
	eraseall(used_hierarchy),
	get_constant(Consts, TopClause, CType, Constant),
	\+ user:NewGoal,
	(NewGoal = OutGoal ; reuse_hierarchy(CType, Constant, NewGoal, OutGoal) ).

reuse_hierarchy(CType, Constant, NewGoal, (H:-B)) :-
	hierarchy_stratified(true),
	found(Constant, NB, CType),
	% avoid repeats
	recordzifnot(used_hierarchy, NB, _),
	functor(NewGoal, Na, 1),
	functor(H, Na, 1),
	functor(B, NB, 1),
	arg(1, H, X),
	arg(1, B, X).


add_consts(Constants, Consts, TopClause, CType, VarClause, Id, Type, TPos, UseAgg, ScoreD0, ScoreF) :-
	choose_best_const(Constants, VarClause, Id, TPos, Type, UseAgg, ScoreD0, ScoreD, _, H, NTPos),
	(  ScoreD0 \= ScoreD
	->
	  debug_format(0,'-DJ- Added to hierarchy  ~w~n',[H]),
	  assert(user:H),
	  H =.. [CName,NewC], 
	  debug_format(2,'-DJ- Looking at  ~w~n',[H]),
	  NewConsts = [NewC|Consts],
	  NewGoal =.. [CName, X],
	  findall(OutGoal,call_constant(NewConsts, TopClause, CType, NewGoal, X, OutGoal),ConstGoals),
	  debug_format(2,'-DJ- Try now adding  ~w~n',[ConstGoals]),
	  add_consts(ConstGoals, NewConsts, TopClause, CType, VarClause, Id, Type, NTPos, UseAgg, ScoreD, ScoreF)
%	  ScoreF = ScoreD
	;
	  ScoreF = ScoreD,
	  debug_format(1,'-DJ- done ~w~n',[ScoreF])
	).

choose_best_const([], _Clause, _Id, TPos, _Type, _UseAgg, Score, Score, H, H, TPos).
choose_best_const([H|Constants], Clause, Id, TPos0, Type, UseAgg, ScoreI, Score, _, HF, TPosF) :-
	assert(user:H),
	score_replacement(Clause, Id, Score1, Type, UseAgg, TPos, TNeg),
	retract(user:H),
%	( H=..[_,V] -> debug_format(1,'-DJ- FAC 4b ~w has ~d/~d ~w~n',[V,TPos,TNeg,Score1]) ;
%	    H=(_:-B), functor(B,V,_), debug_format(1,'-DJ- FAC 4b ~w has ~d/~d ~w~n',[V,TPos,TNeg,Score1]) ),
	% avoid overflows
%	vista_eval_node:minpos(MinPos),
	TPos >= TPos0+TPos0/4,
	vista_eval_node:minacc(MinAcc), TPos/(TPos+TNeg) > MinAcc,
        better_hierarchy_score(Score1, ScoreI), !,
	debug_format(0,'New constant ~w, score ~w:~w/~w~n',[H,Score1,TPos,TNeg]),
	( vista_eval_node:search_mode(eager) ->
	  Score1 = Score, H = HF, TPos = TPosF;	
	  choose_best_const(Constants, Clause, Id, TPos, Type, UseAgg, Score1, Score, H, HF, TPosF) 
	).	
choose_best_const([_|Constants], Clause, Id, TPos, Type, UseAgg, ScoreI, Score, H, HF, TPosF) :-
	choose_best_const(Constants, Clause, Id, TPos, Type, UseAgg, ScoreI, Score, H, HF, TPosF).

complete_search(NewClause, Id, Score) :-
	unvariabilise(NewClause, GClause),
	% finish search expects this in global memory
	replace_clause_in_theory(Id, NewClause, GClause),
	commit_update(Score),
	current_test_score(TestScore),
        %debug_format(1,'PredHead ~w~n', [PredHead]),
	debug_format(1,'~n~n NEW VIEW (score ~w/~w):~n    ~w~n~n',[Score,TestScore,GClause]),
	list_theory.


build_hierarchy(Type) :-
	retractall(hierarchy_ok(all_types)),
	retractall(hierarchy_ok(Type)),
	assert(hierarchy_ok(Type)).

ok_type(_Type) :-
	hierarchy_ok(all_types), !.
ok_type(Type) :-
	hierarchy_ok(Type).

has_hierarchy((Hier,_), Type, Name) :-
	hierarchy_predicate(Hier, Type, Name).

new_hierarchy_predicate_to_call(EmptyPred, Type) :-
	functor(EmptyPred, Name, _),
	assert(hierarchy_predicate(EmptyPred, Type, Name)),
	new_predicate_to_call(EmptyPred).

get_consts(Disj, Name, Consts) :-
	Disj =.. [Name, A],
	findall(A, user:Disj, Consts).

commit_hierarchy(Type,(_:-H,_)) :-
	functor(H,N,1),
	call(user:H),
	arg(1,H,C),
	assert(found(C, N, Type)),
	fail.
commit_hierarchy(_,_).

