
:- module(vista_literals,[all_literals/2]).

:- use_module(modes, [find_modeb/1]).

:- use_module(variables, [create_new_variable/2]).

:- use_module(constants,[get_constants/2]).

:- use_module(library(lists), [append/3]).

%
%
% 
all_literals(Variables, Literals) :-
	findall(X, find_modeb(X), LitList),
	all_literals1(LitList, Variables, Literals).

all_literals1([], _, []).

%gets all literals used with each variables
%does not get all literals (because not backtracking)
all_literals1([Lit | Tail],Variables, Literals) :-
	Lit =.. [Predicate|Args],
    %pass back a list, where each elem is possible instant. of Predicate
	instantiate(Args, Variables, ListOfArgs),
	!,
	plug_in_arguments(Predicate, ListOfArgs, Literals1),
	all_literals1(Tail, Variables, Literals2),
	append(Literals1, Literals2, Literals).

%gets all literals used with each variables
%does not get all literals (because not backtracking)
all_literals1([_ | Tail],Variables, Literals) :-
	all_literals1(Tail, Variables, Literals).

plug_in_arguments(_,[], []).

plug_in_arguments(Predicate, [Head | Tail ], [Lit1 | Rest]) :-
	Lit1 =.. [Predicate | Head],
	plug_in_arguments(Predicate, Tail, Rest).


instantiate([], _, []).

%
instantiate([ArgHead | ArgTail], Variables, ListOfArgs) :-
    instantiateOneArg(ArgHead, Variables, ArgHeadList),
    instantiate(ArgTail, Variables, ArgTailList),
    %append(ArgHeadList, ArgTailList, ListOfArgs).
    distribute(ArgHeadList, ArgTailList, ListOfArgs).

%+ -> 43, - -> 45, # -> 35
% Instantiate one argument of a predicate
% argHead is type
% return the list of variables/constants that are valid input for this argument
instantiateOneArg(ArgHead, Variables, ArgHeadList) :-
    ArgHead = +Type,
    filter_list(Variables, Type, ArgHeadList),
    [] \= ArgHeadList.

instantiateOneArg(ArgHead, Variables, [NewVar | ArgHeadList]) :-
    ArgHead = -Type,
    filter_list(Variables, Type, ArgHeadList),
    create_new_variable(Type, NewVar).

instantiateOneArg(ArgHead, _Variables, ArgHeadList) :-
    ArgHead = '#'(Type),
    get_constants(Type, ArgHeadList).


% filters out the irrelevent variables
filter_list([], _, []).

filter_list([Head|Tail], Type, [Head|ArgHeadList]) :-
    name(Head, [112, 114, 111, 118, 97, 114 | TypeNumber ]),
    get_type(TypeNumber, ThisTypeList),
    name(ThisType, ThisTypeList),
    %write('var type '),
    %write(ThisType),
    %nl,
    Type = ThisType,
    !,
    filter_list(Tail, Type, ArgHeadList).
    
filter_list([_Head|Tail], Type, ArgHeadList) :-
    filter_list(Tail, Type, ArgHeadList).

% gets the type of the variable : given type + number
% input ascii list
% output an ascii list corresponding variable type
% 
get_type([],[]).

%
% In this case we have reached the variable number
%
get_type([Head | _Tail], []) :-
    Head > 47,
    Head < 58,
    !.
%
% Recurse through the type name
%
get_type([Head | Tail], [Head | Rest]) :-
    get_type(Tail, Rest).


%
% distribute and distribute1 find all the possible variablizations of a literal
% base case: Distribute a set of variables over an empty list
% return: a list of lists
%
distribute([H|T],[],[[H]|L]):-
    distribute(T,[],L).

distribute([], _, []).

distribute([H | T], List, ListofLists) :-
    distribute1(H, List, List1),
    distribute(T, List, List2),
    append(List1, List2, ListofLists).

distribute1(_Var1, [], []).

distribute1(Var1, [VarHead | VarList], [[Var1 | VarHead] | ListOfLists]) :-
    distribute1(Var1, VarList, ListOfLists).
    
