
:- module(vista_new_pred,[get_pred_name/1,
			  get_const_pred_name/1,
			  assert_extensionally/3]).

:- use_module(clause_optimise,[optimise_clause/2]).

:- use_module(aggregation,[aggregated/4]).

:- use_module(score,[agg_thresholds/1]).

:- use_module(examples,[call_body/1,
			count/3]).

:- use_module('../common/debug_write',[debug_format/3]).

:- dynamic variable_predicate_count/1,
	constant_predicate_count/1.

variable_predicate_count(1).

constant_predicate_count(1).

get_pred_name(Name) :-
    retract(variable_predicate_count(Num)),
    Num1 is Num +1,
    asserta(variable_predicate_count(Num1)),
    atomic_concat('p', Num, Name).

get_const_pred_name(Name) :-
    retract(constant_predicate_count(Num)),
    Num1 is Num +1,
    asserta(constant_predicate_count(Num1)),
    atomic_concat('hierarchy_', Num, Name).

%
% don't assert extensionally, because we'll soon have bottom-clause.
%
assert_extensionally((H :- Body), aggs, Clause) :-
	functor(H, _, Arity),
	arg(Arity, H, DCount),
	aggregated(Body, H, AggBody, aggs),
	AggBody = aggregate(Count,count,V,B),
	NewBody = count(V, user:B, Count),
	agg_thresholds([Threshold]),
	Clause = (H :- NewBody, (Count > Threshold -> DCount = 1 ; DCount = 0) ),
	assert(user:Clause).
assert_extensionally((H :- Body), single, Clause) :- !,
	Clause = (H :- Body),
	assert(user:Clause).

