
:- module(vista_reduce,[init_reduce/3,
	current_example/2,
	current_seed/3]).

:- use_module(variables, [create_new_variable/2]).

:- use_module(aggregation, [linkable/1,
   linkage/2]).

:- use_module('../common/random',[nextrandom/1]).

:- use_module(newpred,[get_pred_name/1]).

:- use_module(modes,[determination/2,
		     determination/1,
		     determination_arity/1,
		     unassert_modeb/1,
		     modeb/2]).

:- use_module(examples,[next_example/2,random_example/2, example/3]).

:- use_module(foil,[init_foil/1]).

:- use_module('../common/debug_write',[debug_format/3]).

:- use_module(library(lists),[nth0/3, append/3]).

:- dynamic current_example/2.

reduce_tail(Head, PName, Ar) :-
	init_foil(Head),
	functor(EmptyHead,PName,Ar),
	assert(user:new_predicate(EmptyHead)).	

init_reduce(Args, N, single) :-
	start_reduce(Args, N, N, []).
% just add an extra argument at the end.
init_reduce(Args, N, aggs) :-
	N1 is N-1,
	start_reduce(Args, N, N1, [count_t]).

start_reduce(Args, N, N1, TailTypes) :-
	reduce_prelims(N, PName, _, DetAr),
	repeat,
	setof(T,I^linkage(I,T),LL),
	select_randomly(N1, LL, Args, TailTypes),
	create_new_variables(Args, NewVars),
	Head =.. [PName|NewVars],
	generate_mode(Args, DetAr, Modes),
	InMode =.. [PName|Modes],
	functor(S, PName, N),
	modeb(*,InMode),
	( linkable(S) ->  ! ; unassert_modeb(S), fail ),
	debug_format(1,'Learning View ~w~n',[InMode]),
	reduce_tail(Head, PName, N).

reduce_prelims(Arity, PName, _PNum, Ar) :-
	set_value(processed_clauses,0),
	set_value(refinements,0),
	eraseall(clause),
	eraseall(to_refine),
	eraseall(fast_clause),
	retractall(current_example(_,_)),
	next_example(Id, pos),
	assert(current_example(Id,pos)),
	retractall(user:new_predicate(_)),
	get_pred_name(PName),
	determination(S), functor(S,A,Ar),
	determination(A/Ar,PName/Arity).

current_seed(ExId, PosNeg, Ex) :-
	current_example(ExId,PosNeg),
	example(ExId, PosNeg, Ex).

choose_type(Arg1,L) :-
	nextrandom(X),
	length(L, Size),
	Pos is integer(Size * X),
	nth0(Pos, L, Ts),
	Ts =.. [_|LTs],
	select_random(LTs, Arg1).

select_random([], []).
select_random(Types.L, Type.LTs) :-
	select_random_type(Types, Type),
	select_random(L, LTs).

select_randomly(0,_) --> !.
select_randomly(I0,Types) -->
	{
         select_random_type(Types, T),
	 I is I0-1
        },
	[T],
	select_randomly(I,Types).

select_random_type(Types, Type) :-
	nextrandom(X),
	length(Types, Size),
	Pos is integer(Size * X),
	nth0(Pos, Types, Type).

generate_mode([], _, []).
generate_mode([count_t], _, [-count_t]) :- !.
generate_mode([Type|KeyTypes], 0, [-Type|Modes]) :-
	generate_mode(KeyTypes, 0, Modes).
generate_mode([Type|KeyTypes], N, [+Type|Modes]) :-
	N1 is N-1,
	generate_mode(KeyTypes, N1, Modes).

create_new_variables([], []).
create_new_variables(Arg.Args, NewArg.NewArgs) :-
        create_new_variable(Arg, NewArg),
        create_new_variables(Args, NewArgs).
