
:- module(vista_score_node,[score_clause/6,
			    score_replacement/7,
			    counts/4,
			    assert_counts/4,
			    agg_thresholds/1]).

:- use_module('../common/debug_write',[debug_format/3,
				       debug_do_and_format/4
				      ]).

:- use_module('../common/utils',[average2/3
				      ]).

:- use_module(clause_optimise,[optimise_clause/2]).

:- use_module(examples,[call_example/4,
			call_agg/7]).

:- use_module('../scorer/scorer',[new_feature/5,
				  replace_feature/6,
	                          scorer_has_tuneset/0,
				  scoring_data/4,
				  covers_example/4,
				  init_scoring/1,
				  vals_to_scorable/3,
				  count_cover/2]).

:- dynamic counts/4.

:- dynamic agg_thresholds/1.

agg_thresholds([]).

assert_counts(_,_,_,_) :-
	retractall(counts(_,_,_,_)),
	fail.				       
assert_counts(TrPCounts,TrNCounts,TePCounts,TeNCounts) :-
	assert(counts(TrPCounts,TrNCounts,TePCounts,TeNCounts)).

%
% first, generate a bitmaps as a list of 0 and 1s, then send it off to Java.
%
score_clause(Clause, Scores, _Type, UseAggregates, TPos, TNeg) :-
%	format(user_error,'-DJ- FAC 3a ~w~n',[UseAggregates]),
	compute_cover(Clause, UseAggregates, TrPos, TrNeg, TePos, TeNeg, TPos, TNeg),
%	format(user_error,'-DJ- FAC 3a ~w~n',[TPos:TNeg]),
	new_feature(TrPos,TrNeg,TePos,TeNeg, Scores).
%	format(user_error,'[ ~w/~w ]~n',[TPos,Scores]),
%	portray_clause(Clause).

score_replacement(Clause, -1, Scores, Type, UseAggregates, TPos,TNeg) :- !,
%	format(user_error,'-DJ- FAC 3a ~w~n',[UseAggregates]),
	score_clause(Clause, Scores, Type, UseAggregates, TPos, TNeg).
score_replacement(Clause, Id, Scores, _Type, UseAggregates, TPos, TNeg) :-
%	format(user_error,'-DJ- FAC 3a ~w~n',[UseAggregates]),
	compute_cover(Clause, UseAggregates, TrP, TrN, TeP, TeN, TPos, TNeg),
	replace_feature(Id,TrP,TrN,TeP,TeN,Scores).


compute_cover(Clause, aggs, TrP, TrN, TeP, TeN, TPos, TNeg) :-
	Clause = (_ :- _), !,
	format(user_error,'-DJ- FAC 3a ~w~n',[Clause]),
	(
	    scorer_has_tuneset
	->
   	   %format(user_error,'-DJ- FAC 3b~n',[]),
	   gen_cover(Clause, train, pos, TrainPosCover, TunePosCover),
	   %format(user_error,'-DJ- FAC 3c~n',[]),
	   gen_cover(Clause, train, neg, TrainNegCover, TuneNegCover),
	   gen_cover(Clause, tune, pos, TunePosCover, []),
	   %format(user_error,'-DJ- FAC 3c~n',[]),
	   gen_cover(Clause, tune, neg, TuneNegCover, [])
        ;
   	   %format(user_error,'-DJ- FAC 3b~n',[]),
	   gen_cover(Clause, train, pos, TrP, []),
	   %format(user_error,'-DJ- FAC 3c~n',[]),
	   gen_cover(Clause, train, neg, TrN, [])
        ),
	average2(TrainPosCover, TrainNegCover, Avg),
	retractall(agg_thresholds(_)),
	assert(agg_thresholds([Avg])),
	scoring_data(TrP,TrN,TeP,TeN),
writeln(TrP:TrN),
	init_scoring(TrP),
	init_scoring(TrN),
	binarise(TrainPosCover, pos, test, 1, Avg, TrP),
	binarise(TrainNegCover, neg, test, 1, Avg, TrN),
	gen_cover(Clause, test, pos, TestPosCover, []),
	gen_cover(Clause, test, neg, TestNegCover, []),
writeln(TeP:TeN),
	init_scoring(TeP),
	init_scoring(TeN),
	binarise(TestPosCover, pos, test, 1, Avg, TeP),
	binarise(TestNegCover, neg, test, 1, Avg, TeN),
	%format(user_error,'-DJ- FAC 3d~n',[]),
	count_cover(TrP,TPos),
	count_cover(TrN,TNeg),
	debug_format(2,'  scores: ~d/~d~n', [TPos, TNeg]).
% also handles base case for aggs.
compute_cover(Clause, _, TrP, TrN, TeP, TeN, TPos, TNeg) :-
	(Clause = (_ :- B) -> true ; B = true ),
	scoring_data(TrP,TrN,TeP,TeN),
	(
	    scorer_has_tuneset
	->
	   TrP = TrPTr-TrPTu,
	   TrN = TrNTr-TrNTu,
   	   %format(user_error,'-DJ- FAC 3b~n',[]),
	   gen_bmap(Clause, train, pos, TrPTr),
	   %format(user_error,'-DJ- FAC 3c~n',[]),
	   gen_bmap(Clause, train, neg, TrNTr),
	   gen_bmap(Clause, tune, pos, TrPTu),
	   %format(user_error,'-DJ- FAC 3c~n',[]),
	   gen_bmap(Clause, tune, neg, TrNTu)
        ;
   	   %format(user_error,'-DJ- FAC 3b~n',[]),
	   gen_bmap(Clause, train, pos, TrP),
	   %format(user_error,'-DJ- FAC 3c~n',[]),
	   gen_bmap(Clause, train, neg, TrN)
        ),
	gen_bmap(Clause, test, pos, TeP),
	gen_bmap(Clause, test, neg, TeN),
	%format(user_error,'-DJ- FAC 3d~n',[]),
	count_cover(TrP,TPos),
	count_cover(TrN,TNeg),
	debug_format(2,'  scores: ~d/~d~n', [TPos, TNeg]).

gen_bmap(Ex, Set, PosNeg, Obj) :-
	score_simple_rule(Set, PosNeg, Ex, Obj).

score_simple_rule(Type, PosNeg, Clause, Obj) :-
	init_scoring(Obj),
	call_example(Type, PosNeg, Clause, Obj).
score_simple_rule(_, _, _, Obj) :-
	debug_do_and_format(3,jpl_array_to_list(Obj, Scores),
	     'Score ~w~n',[Scores]).
	

gen_cover((Ex :- B), Set, PosNeg, Vals, MoreVals) :- !,
	%format(user_error,'-DJ- FAC 3b1~n',[]),
	split_body_aggs(B, Linkage, Body, OV, V),
	optimise_clause((Ex :- Body), (_ :- NBody)),
	NBody \= '!'(_,_),
	findall(V,call_agg(OV,Set,PosNeg,Ex,Linkage,NBody,V),Vals, MoreVals).

split_body_aggs(aggregate(V,_,OV,Body),true,Body,OV,V) :- !.
split_body_aggs((G,Body),(G,Linkage),B,OV,V) :-
	split_body_aggs(Body,Linkage,B,OV,V).

binarise([], _, _, _, _, _).
binarise(V.Vals, PosNeg, Set, I0, Avg, Obj) :-
	I is I0+1,
	( V > Avg -> 
	    covers_example(Obj, PosNeg, Set, I0)
	;
	   true
	),
	binarise(Vals, PosNeg, Set, I, Avg, Obj).

