
:- module(vista_valid,[valid_clause/1,
		       strict_valid_clause/1]).

:- use_module(goals,[diff_goals/3]).

:- use_module(library(terms),[variable_in_term/2]).

valid_clause((Head :- Body)) :-
	Head =..[_ | Args],
	diff_goals(Body,[],_),
	check_a_variable(Args, Body).

strict_valid_clause((Head :- Body)) :-
	Head =..[_ | Args],
	check_variable(Args, Body),
	number_goals_and_get_vars(Body,[],1,_,[],Conj),
	calculate_independent_sets(Conj,[],[[_|_]]), !.

check_variable([],_).
check_variable([Var|T], Body) :-
    variable_in_term(Body, Var),
    check_variable(T, Body).

check_a_variable([Var|_], Body) :-
    variable_in_term(Body, Var), !.
check_a_variable([_|T], Body) :-
    check_a_variable(T, Body).


number_goals_and_get_vars((G,Body),LHead,I0,IF,L0,LGs) :- !,
	number_goals_and_get_vars(G,LHead,I0,I1,L0,LI),
	number_goals_and_get_vars(Body,LHead,I1,IF,LI,LGs).
number_goals_and_get_vars(G,LHead,I0,I,L0,[g(I,LVF,NG)|L0]) :-
	I is I0+1,
	get_goal_vars(G,LHead,LVF,NG).

get_goal_vars(G,LHead,LVF,G) :-
	term_variables(G,LVI0),
	sort(LVI0, LVI),
%	vars_in_term([G],[],LVI),
	ord_subtract(LVI,LHead,LVF).

calculate_independent_sets([],BSets,BSets).
calculate_independent_sets([G|Ls],BSets0,BSetsF) :-
	add_goal_to_set(G,BSets0,BSetsI),
	calculate_independent_sets(Ls,BSetsI,BSetsF).

add_goal_to_set(g(I,LV,G),Sets0,SetsF) :-
	add_to_sets(Sets0,LV,[g(I,LV,G)],SetsF).

add_to_sets([],LV,Gs,[[LV|Gs]]).
add_to_sets([[LV|Gs]|Sets0],LVC,GsC,[[LV|Gs]|SetsF]) :-
	ord_disjoint(LV,LVC), !,
	add_to_sets(Sets0,LVC,GsC,SetsF).
add_to_sets([[LV|Gs]|Sets0],LVC,GsC,SetsF) :-
	ord_union(LV,LVC,LVN),
	join_goals(Gs,GsC,GsN),
	add_to_sets(Sets0,LVN,GsN,SetsF).

join_goals([],L,L):- !.
join_goals(L,[],L):- !.
join_goals([g(I1,VL1,G1)|T],[g(I2,VL2,G2)|T2],Z) :-
	I1 < I2, !,
	Z = [g(I1,VL1,G1)|TN],
	join_goals(T,[g(I2,VL2,G2)|T2],TN).
join_goals([H|T],[g(I2,VL2,G2)|T2],Z) :-
	Z = [g(I2,VL2,G2)|TN],
	join_goals(T,[H|T2],TN).


